from environmentbase.template import Template
from troposphere import Ref, Output, GetAtt, ec2

SSH_PORT = '22'


class Bastion(Template):
    """
    Adds a bastion host within a given deployment based on environemntbase.
    """

    def __init__(self, name='bastion', ingress_port='2222', access_cidr='0.0.0.0/0', instance_type='t2.micro', user_data=None):
        """
        Method initializes bastion host in a given environment deployment
        @param name [string] - name of the tier to assign
        @param ingress_port [number] - port to allow ingress on. Must be a valid ELB ingress port. More info here:
        http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-elb-listener.html
        @param access_cidr [string] - CIDR notation for external access to this tier.
        @param user_data [string] - User data to to initialize the bastion hosts.
        """

        self.name = name
        self.ingress_port = ingress_port
        self.access_cidr = access_cidr
        self.instance_type = instance_type
        self.user_data = user_data

        super(Bastion, self).__init__(template_name=name)

    # Called after add_child_template() has attached common parameters and some instance attributes:
    # - RegionMap: Region to AMI map, allows template to be deployed in different regions without updating AMI ids
    # - ec2Key: keyname to use for ssh authentication
    # - vpcCidr: IP block claimed by whole VPC
    # - vpcId: resource id of VPC
    # - commonSecurityGroup: sg identifier for common allowed ports (22 in from VPC)
    # - utilityBucket: S3 bucket name used to send logs to
    # - [public|private]Subnet[0-9]: indexed and classified subnet identifiers
    #
    # and some instance attributes referencing the attached parameters:
    # - self.vpc_cidr
    # - self.vpc_id
    # - self.common_security_group
    # - self.utility_bucket
    # - self.subnets: keyed by type and index (e.g. self.subnets['public'][1])
    # - self.user_data: User data for launch configuration
    def build_hook(self):
        """
        Hook to add tier-specific assets within the build stage of initializing this class.
        """
        security_groups = self.add_security_groups()

        bastion_elb = self.add_elb(
            resource_name=self.name,
            security_groups=[security_groups['elb']],
            listeners=[{
                'elb_port': self.ingress_port,
                'instance_port': SSH_PORT
            }],
            utility_bucket=self.utility_bucket
        )

        self.add_asg(
            layer_name=self.name,
            security_groups=[security_groups['bastion'], self.common_security_group],
            load_balancer=bastion_elb,
            user_data=self.user_data,
            instance_type=self.instance_type
        )

        self.add_output(Output(
            'BastionELBDNSName',
            Value=GetAtt(bastion_elb, 'DNSName')
        ))

        self.add_output(
            Output(
                'BastionELBDNSZoneId',
                Value=GetAtt(bastion_elb, 'CanonicalHostedZoneNameID')))

        self.add_output(Output(
            'BastionSecurityGroupId',
            Value=Ref(security_groups['bastion'])
        ))

    @staticmethod
    def get_factory_defaults():
        return {"bastion": {
            "instance_type": "t2.micro",
            "remote_access_cidr": "0.0.0.0/0",
            "ingress_port": 2222
        }}

    @staticmethod
    def get_config_schema():
        return {"bastion": {
            "instance_type": "str",
            "remote_access_cidr": "str",
            "ingress_port": "int"
        }}

    def add_security_groups(self):
        """
        Wrapper method to encapsulate process of creating security groups for this tier.
        """

        elb_sg_ingress_rule = ec2.SecurityGroupRule(
            FromPort=self.ingress_port,
            ToPort=self.ingress_port,
            IpProtocol='tcp',
            CidrIp=self.access_cidr)

        elb_sg_name = '%sElbSecurityGroup' % self.name
        elb_sg = self.add_resource(
            ec2.SecurityGroup(
                elb_sg_name,
                GroupDescription='Security group for %s ELB' % self.name,
                VpcId=self.vpc_id,
                SecurityGroupIngress=[elb_sg_ingress_rule])
        )

        bastion_sg_name = '%sSecurityGroup' % self.name
        bastion_sg = self.add_resource(
            ec2.SecurityGroup(
                bastion_sg_name,
                GroupDescription='Security group for %s' % self.name,
                VpcId=self.vpc_id)
        )

        self.create_reciprocal_sg(
            elb_sg, elb_sg_name,
            bastion_sg, bastion_sg_name,
            from_port=SSH_PORT)

        return {'bastion': bastion_sg, 'elb': elb_sg}



# -*- coding: utf-8 -*-
import re
import os
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinxcontrib.napoleon'
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = u'cfn-environment-base'
copyright = u'2017, Patrick Pierson'
version = release =  re.findall(
    "__version__ = '(.*)'", 
    open(os.path.join(os.path.dirname('../'), 'src/environmentbase/version.py')).read()
)[0]

import sphinx_py3doc_enhanced_theme
html_theme = "sphinx_py3doc_enhanced_theme"
html_theme_path = [sphinx_py3doc_enhanced_theme.get_html_theme_path()]

pygments_style = 'trac'
templates_path = ['.']
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = True
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)
html_theme_options = {
    'githuburl': 'https://github.com/AWSFrederick/cloudformation-environmentbase/'
}



#!/usr/bin/env python
'''Generate parameters for given template with 
values from prereq_stack(s) that have already been deployed.

Usage:
    ./get_parameters.py <s3_root_template> [<prereq_stacks>...] [--extra-yaml=FILE...] [--output=cli|json]

Options:
    -h --help               Show this screen.
    -v --version            Show version.
    s3_root_template        Like: "s3://tropos-bucket/templates/v2/microservices-root.template"
    --output=<format>       Specify output format [default: json]

Examples: 
    python get_parameters.py s3://tropos-bucket/templates/v2/microservices-root.template networkbase --output=cli
    python get_parameters.py s3://tropos-bucket/templates/v2/microservices-root.template networkbase user-runtime-root  --output=cli
    $ python get_parameters.py s3://tropos-bucket/templates/v2/microservices-root.template networkbase user-runtime-root | jq .
    [
      {
        "ParameterValue": "sg-bc973edb",
        "ParameterKey": "SwarmManagerSecurityGroupId"
      }, [...]
    ]
    $ python get_parameters.py s3://tropos-bucket/templates/v2/microservices-root.template networkbase user-runtime-root --output cli
        ParameterKey=SwarmManagerSecurityGroupId,ParameterValue=sg-bc973edb ...
'''
import urlparse
import logging
import json
import yaml

from docopt import docopt
import boto3

logging.basicConfig()
logger = logging.getLogger('get_parameters.py')
# logger.setLevel(logging.DEBUG)

CLI_FORMAT = 'ParameterKey={key},ParameterValue={value}'
JSON_FORMAT = {'ParameterKey': None, 'ParameterValue': None}

client_s3 = boto3.client('s3')
client_cfn = boto3.client('cloudformation')

def main(arguments):
    s3_root_template = arguments['<s3_root_template>']
    prereq_stacks = arguments['<prereq_stacks>']
    output_format = arguments['--output']
    extra_yamls = arguments['--extra-yaml']

    parameters = get_root_params(s3_root_template)
    additional_outputs = get_additional_outputs(prereq_stacks)
    additional_parameters = get_additional_parameters(extra_yamls)
    additional_outputs.update(additional_parameters)
    final_parameters = dict(get_final_parameters(parameters, additional_outputs))
    print format_final_parameters(final_parameters, output_format)

def get_root_params(s3_root_template):
    s3, bucket, key, _, _, _ = urlparse.urlparse(s3_root_template)
    key = key.strip('/')
    logger.debug( (bucket, key) )
    response = client_s3.get_object(Bucket=bucket, Key=key)
    body = response['Body']
    template = json.load(body)
    parameters = template.get('Parameters')
    logger.debug( parameters )
    return parameters


def get_additional_outputs(prereq_stacks):
    response = client_cfn.describe_stacks()
    stacks = response['Stacks']

    additional_outputs = {}
    for stack in stacks:
        if stack['StackName'] not in prereq_stacks:
            continue
        stack_outputs = stack.get('Outputs')
        stack_outputs = {d['OutputKey']: d['OutputValue'] for d in stack_outputs}
        additional_outputs.update(stack_outputs)

    logger.debug( additional_outputs )
    return additional_outputs


def get_additional_parameters(extra_yamls):
    additional_parameters = {}
    for param_file in extra_yamls:
        with open(param_file) as f:
            additional_parameters.update(yaml.load(f))
    return additional_parameters


def get_final_parameters(parameters, additional_outputs):
    # ParameterKey=KeyPairName, ParameterValue=MyKey
    # {'ParameterKey': key, 'ParameterValue': additional_outputs[key]}
    for key in parameters:
        if key in additional_outputs:
            yield key, additional_outputs[key]


def format_final_parameters(final_parameters, output_format='json'):
    if output_format == 'cli':
        return ' '.join([CLI_FORMAT.format(key=key, value=final_parameters[key]) 
            for key in final_parameters])
    else:
        return json.dumps([{'ParameterKey': key, 'ParameterValue': final_parameters[key]} 
            for key in final_parameters])


if __name__ == '__main__':
    arguments = docopt(__doc__, version='0.1')
    logger.info((arguments) )
    main(arguments)



__author__ = 'Eric Price'

from datetime import date
import boto3
import json
import re


VOL_TYPE_MAG = 'standard'
VOL_TYPE_SSD = 'gp2'

VIRT_TYPE_PV = 'paravirtual'
VIRT_TYPE_HVM = 'hvm'

ROOT_DEV_EBS = 'ebs'
ROOT_DEV_INS = 'instance-store'

YEAR = date.today().year

#  amzn-ami-hvm-2015.03.rc-0.x86_64-gp2
RC_REGEX = re.compile(r'rc-[0-9]')

# amzn-ami-vpc-nat-hvm-2014.03.2.x86_64-gp2
NAT_REGEX = re.compile(r'-vpc-nat-')

# amzn-ami-minimal-pv-2014.03.2.x86_64-s3
MINIMAL_REGEX = re.compile(r'-minimal-')

# aws ec2 describe-images --owners amazon --filters Name=name,Values="amzn-ami-hvm-2015*-gp2"


def get_region_list():
    client = boto3.client('ec2')
    response = client.describe_regions()

    return_code = response['ResponseMetadata']['HTTPStatusCode']
    if return_code != 200:
        raise Exception('describe_regions returned %s' % return_code)
    else:
        return map(lambda entry: entry['RegionName'], response['Regions'])


def find_amis(region='us-west-2', year=YEAR):
    client = boto3.client('ec2', region_name=region)

    ami_name = 'amzn-ami-*-%s.*.x86_64-*' % year
    filter_name = [
        {'Name': 'name', 'Values': [ami_name]},
        {'Name': 'root-device-type', 'Values': [ROOT_DEV_EBS]},  # no instance-store images
        {'Name': 'virtualization-type', 'Values': [VIRT_TYPE_PV, VIRT_TYPE_HVM]},
    ]
    response = client.describe_images(
        Owners=['amazon'],
        Filters=filter_name)

    return_code = response['ResponseMetadata']['HTTPStatusCode']
    if return_code != 200:
        raise Exception('describe_image returned %s' % return_code)
    else:
        return response['Images']


def filter_amis(ami_list, remove_rcs=True, use_nat_instances=False, remove_minimal=True):
    #  Remove release candidates
    if remove_rcs:
        ami_list = filter(lambda ami: RC_REGEX.search(ami['Name']) is None, ami_list)

    # Looking for NAT instances?
    if use_nat_instances:
        nat_filter_fun = lambda ami: NAT_REGEX.search(ami['Name']) is not None
    else:
        nat_filter_fun = lambda ami: NAT_REGEX.search(ami['Name']) is None
    ami_list = filter(nat_filter_fun, ami_list)

    # Filter out 'minimal' instance types
    if remove_minimal:
        ami_list = filter(lambda ami: MINIMAL_REGEX.search(ami['Name']) is None, ami_list)

    if len(ami_list) == 0:
        return None

    ami_list = sorted(ami_list, key=lambda ami: ami['CreationDate'], reverse=True)

    return ami_list


def get_hvm_ami(filtered_amis, vol_type=VOL_TYPE_SSD):
    filter_fun = lambda ami: ami['VirtualizationType'] == VIRT_TYPE_HVM and ami['BlockDeviceMappings'][0]['Ebs']['VolumeType'] == vol_type
    return filter(filter_fun, filtered_amis)[0]


def get_pv_ami(filtered_amis):
    filter_fun = lambda ami: ami['VirtualizationType'] == VIRT_TYPE_PV \
        and ami['BlockDeviceMappings'][0]['Ebs']['VolumeType'] == VOL_TYPE_MAG
    return filter(filter_fun, filtered_amis)[0]


if __name__ == '__main__':
    region_mapping = {}
    regions = get_region_list()

    print('Regions %s' % regions)
    for region in regions:
        print('\n-------------------\nProcessing %s' % region)
        images = find_amis(region, YEAR)
        filtered_amis = filter_amis(images)

        # If no hits retry with last years AMIs
        if filtered_amis is None:
            images = find_amis(region, YEAR - 1)
            filtered_amis = filter_amis(images)

        hvm_ami = get_hvm_ami(filtered_amis)
        print(json.dumps(hvm_ami, indent=4, separators=(',', ': ')))

        pv_ami = get_pv_ami(filtered_amis)
        print(json.dumps(pv_ami, indent=4, separators=(',', ': ')))

        region_mapping[region] = {'PV64': pv_ami["ImageId"], 'HVM64': hvm_ami["ImageId"]}

    json_str = json.dumps(region_mapping, indent=4, separators=(',', ': '), sort_keys=True)

    # Someone please explain to me why I have to do this repeatedly for all matches to be replaced!!
    import re
    for _ in range(7):
        json_str = re.sub(r": \{\n\s*([^\n]+),\n\s*([^\n]+)\n\s*\}", r": { \1, \2 }", json_str, re.MULTILINE)

    print('\n========================')
    print(json_str)



from troposphere import Ref, FindInMap, Output, GetAZs, Select
import troposphere.ec2 as ec2
from . import ha_nat
import netaddr
from toolz import groupby, assoc

from environmentbase.template import Template


class BaseNetwork(Template):
    DEFAULT_CONFIG = {
        "network": {
            "network_cidr_base": "10.0.0.0",
            "network_cidr_size": "16",
            "az_count": 3,
            "subnet_types": [
                "public",
                "private"
            ],
            "subnet_config": [
                {
                    "type": "public",
                    "size": "18",
                    "name": "public"
                },
                {
                    "type": "private",
                    "size": "22",
                    "name": "private"
                },
            ],
        },
        "nat": {
            "instance_type": "t2.micro",
            "enable_ntp": False
        }
    }

    CONFIG_SCHEMA = {
        "network": {
            "az_count": "int",
            "subnet_types": "list",
            "subnet_config": "list",
            "network_cidr_base": "basestring",
            "network_cidr_size": "basestring"
        },
        "nat": {
            "instance_type": "basestring",
            "enable_ntp": "bool"
        }
    }

    # When no config.json file exists a new one is created using the 'factory default' file.  This function
    # augments the factory default before it is written to file with the config values required
    @staticmethod
    def get_factory_defaults():
        return BaseNetwork.DEFAULT_CONFIG

    # When the user request to 'create' a new BaseNetwork template the config.json file is read in. This file is checked to
    # ensure all required values are present. Because BaseNetwork has additional requirements beyond that of
    # EnvironmentBase this function is used to add additional validation checks.
    @staticmethod
    def get_config_schema():
        return BaseNetwork.CONFIG_SCHEMA

    def __init__(self, template_name, network_config, nat_config):
        super(BaseNetwork, self).__init__(template_name)

        self.network_config = network_config
        self.nat_config = nat_config
        self.az_count = int(network_config.get('az_count', '2'))
        self.stack_outputs = {}

        # Simple mapping of AZs to NATs, to prevent creating duplicates
        self.az_nat_mapping = {}

        self._igw = None
        self._vpc_gateway_attachment = None

        self.construct_network()

    @property
    def igw(self):
        return self._ref_maybe(self._igw)

    @property
    def vpc_gateway_attachment(self):
        return self._ref_maybe(self._vpc_gateway_attachment)

    def build_hook(self):
        # Remove the common parameters that this stack creates
        for param_name in ["igwVpcAttachment", "vpcCidr", "internetGateway", "commonSecurityGroup", "vpcId"]:
            if param_name in self.parameters:
                self.parameters.pop(param_name)

    def construct_network(self):
        """
        Main function to construct VPC, subnets, security groups, NAT instances, etc
        """
        network_config = self.network_config
        nat_config = self.nat_config
        # az_count = self.az_count

        self.add_network_cidr_mapping(network_config=network_config)
        self._prepare_subnets(self._subnet_configs)
        self.create_network_components(network_config=network_config, nat_config=nat_config)

        self._common_security_group = self.add_resource(
            ec2.SecurityGroup(
                'commonSecurityGroup',
                GroupDescription='Security Group allows ingress and egress for common usage patterns throughout this infrastructure.',
                VpcId=self.vpc_id,
                SecurityGroupEgress=[
                    ec2.SecurityGroupRule(
                        FromPort='80',
                        ToPort='80',
                        IpProtocol='tcp',
                        CidrIp='0.0.0.0/0'),
                    ec2.SecurityGroupRule(
                        FromPort='443',
                        ToPort='443',
                        IpProtocol='tcp',
                        CidrIp='0.0.0.0/0'),
                    ec2.SecurityGroupRule(
                        FromPort='123',
                        ToPort='123',
                        IpProtocol='udp',
                        CidrIp='0.0.0.0/0')],
                SecurityGroupIngress=[
                    ec2.SecurityGroupRule(
                        FromPort='22',
                        ToPort='22',
                        IpProtocol='tcp',
                        CidrIp=FindInMap('networkAddresses', 'vpcBase', 'cidr'))]))

        self.add_output(Output('commonSecurityGroup', Value=self.common_security_group))

    def _prepare_subnets(self, subnet_configs):
        for index, subnet_config in enumerate(subnet_configs):
            subnet_type = subnet_config.get('type', 'private')
            subnet_layer = subnet_config.get('name', 'subnet')
            subnet_az = subnet_config.get('AZ', '-1')

            subnet_name = subnet_layer + 'AZ' + str(subnet_az)

            # Save the subnet references to the template object
            if subnet_type not in self._subnets:
                self._subnets[subnet_type] = {}

            if subnet_layer not in self._subnets[subnet_type]:
                self._subnets[subnet_type][subnet_layer] = []

            self._subnets[subnet_type][subnet_layer].append(Ref(subnet_name))

    def create_network_components(self, network_config, nat_config):
        """
        Method creates a network with the specified number of public and private subnets within the
        VPC cidr specified by the networkAddresses CloudFormation mapping.
        @param network_config [dict] collection of network parameters for creating the VPC network
        """

        # make VPC
        if 'network_name' in network_config:
            network_name = network_config.get('network_name')
        else:
            network_name = self.__class__.__name__

        self._vpc_cidr = FindInMap('networkAddresses', 'vpcBase', 'cidr')
        self.add_output(Output('networkAddresses', Value=str(self.mappings['networkAddresses'])))
        self.add_output(Output('vpcCidr', Value=self.vpc_cidr))

        self._vpc_id = self.add_resource(
            ec2.VPC(
                'vpc',
                CidrBlock=self._vpc_cidr,
                EnableDnsSupport=True,
                EnableDnsHostnames=True,
                Tags=[ec2.Tag(key='Name', value=network_name)]))

        self.add_output(Output('vpcId', Value=self.vpc_id))

        self._igw = self.add_resource(ec2.InternetGateway('vpcIgw'))
        self.add_output(Output('internetGateway', Value=self.igw))

        # add IGW
        igw_title = 'igwVpcAttachment'
        self._vpc_gateway_attachment = self.add_resource(ec2.VPCGatewayAttachment(
            igw_title,
            InternetGatewayId=self.igw,
            VpcId=self.vpc_id))

        self.add_output(Output('igwVpcAttachment', Value=self.vpc_gateway_attachment))

        self.gateway_hook()

        # make Subnets
        for index, subnet_config in enumerate(self._subnet_configs):
            subnet_type = subnet_config.get('type', 'private')
            # subnet_size = subnet_config.get('size', '22')
            subnet_layer = subnet_config.get('name', 'subnet')
            subnet_az = subnet_config.get('AZ', '-1')
            subnet_cidr = subnet_config.get('cidr', 'ERROR')
            # az_key = 'AZ{}'.format(subnet_az)

            CidrBlock = subnet_cidr
            # Create the subnet
            subnet_name = subnet_layer + 'AZ' + str(subnet_az)
            subnet = self.add_resource(ec2.Subnet(
                subnet_name,
                AvailabilityZone=Select(subnet_az, GetAZs()),
                VpcId=self.vpc_id,
                CidrBlock=CidrBlock,
                Tags=[ec2.Tag(key='network', value=subnet_type),
                      ec2.Tag(key='Name', value=subnet_name)]))

            self.add_output(Output(subnet_name, Value=self._ref_maybe(subnet)))

            # Create the routing table
            route_table = self.add_resource(ec2.RouteTable(
                subnet_name + 'RouteTable',
                VpcId=self.vpc_id))

            # Create the NATs and egress rules
            self.create_subnet_egress(subnet_az, route_table, igw_title, subnet_type, subnet_layer, nat_config)

            # Associate the routing table with the subnet
            self.add_resource(ec2.SubnetRouteTableAssociation(
                subnet_name + 'EgressRouteTableAssociation',
                RouteTableId=Ref(route_table),
                SubnetId=Ref(subnet)))

    def create_subnet_egress(self, subnet_az, route_table, igw_title, subnet_type, subnet_layer, nat_config):
        """
        Create an egress route for the subnet with the given subnet_az and type
        Override to create egress routes for other subnet types
        Creates the NAT instances in the public subnets
        """

        # For public subnets, create the route to the IGW
        if subnet_type == 'public':
            self.add_resource(
                ec2.Route(
                    subnet_layer + 'AZ' + str(subnet_az) + 'EgressRoute',
                    DependsOn=[igw_title],
                    DestinationCidrBlock='0.0.0.0/0',
                    GatewayId=self.igw,
                    RouteTableId=Ref(route_table)))

        # For private subnets, create a NAT instance in a public subnet in the same AZ
        elif subnet_type == 'private':

            # If we have already created a NAT in this AZ, skip it
            if self.az_nat_mapping.get(subnet_az):
                return

            nat_instance_type = nat_config['instance_type']
            nat_enable_ntp = nat_config['enable_ntp']
            extra_user_data = nat_config.get('extra_user_data')
            ha_nat = self.create_nat(
                subnet_az,
                nat_instance_type,
                nat_enable_ntp,
                name='HaNat' + str(subnet_az),
                extra_user_data=extra_user_data)

            # We merge the NAT template into the root template
            self.merge(ha_nat)

            # Save the reference to the HA NAT, so we don't recreate it if we hit another private subnet in this AZ
            self.az_nat_mapping[subnet_az] = ha_nat

    def gateway_hook(self):
        """
        Override to allow subclasses to create VPGs and similar components during network creation
        """
        pass

    def create_nat(self, index, nat_instance_type, enable_ntp, name, extra_user_data=None):
        """
        Override to customize your NAT instance. The returned object must be a
        subclass of ha_nat.HaNat.
        """
        return ha_nat.HaNat(
            index,
            nat_instance_type,
            enable_ntp,
            name=name,
            extra_user_data=extra_user_data)

    def _get_subnet_config_w_az(self, network_config):
        az_count = int(network_config.get('az_count', 2))
        subnet_config = network_config.get('subnet_config', {})

        for subnet in subnet_config:
            for az in range(az_count):
                newsubnet = assoc(subnet, 'AZ', az)
                yield newsubnet

    def _get_subnet_config_w_cidr(self, network_config):
        network_cidr_base = str(network_config.get('network_cidr_base', '172.16.0.0'))
        network_cidr_size = str(network_config.get('network_cidr_size', '20'))
        # first_network_address_block = str(network_config.get('first_network_address_block', network_cidr_base))

        # ret_val = {}
        base_cidr = network_cidr_base + '/' + network_cidr_size
        net = netaddr.IPNetwork(base_cidr)

        grouped_subnet = groupby('size', self._get_subnet_config_w_az(network_config))
        subnet_groups = sorted(grouped_subnet.items())
        available_cidrs = []

        for subnet_size, subnet_configs in subnet_groups:
            newcidrs = net.subnet(int(subnet_size))

            for subnet_config in subnet_configs:
                try:
                    cidr = newcidrs.next()
                except StopIteration as e:
                    net = chain(*reversed(available_cidrs)).next()
                    newcidrs = net.subnet(int(subnet_size))
                    cidr = newcidrs.next()

                new_config = assoc(subnet_config, 'cidr', str(cidr))
                yield new_config
            else:
                net = newcidrs.next()
                available_cidrs.append(newcidrs)

    def add_network_cidr_mapping(self, network_config):
        """
        Method calculates and adds a CloudFormation mapping that is used to set VPC and Subnet CIDR blocks.
        Calculated based on CIDR block sizes and additionally checks to ensure all network segments
        fit inside of the specified overall VPC CIDR.
        @param network_config [dict] dictionary of values containing data for creating
        """
        # az_count = int(network_config.get('az_count', '2'))
        network_cidr_base = str(network_config.get('network_cidr_base', '172.16.0.0'))
        network_cidr_size = str(network_config.get('network_cidr_size', '20'))
        # first_network_address_block = str(network_config.get('first_network_address_block', network_cidr_base))

        ret_val = {}
        base_cidr = network_cidr_base + '/' + network_cidr_size
        # cidr_info = Network(base_cidr)
        ret_val['vpcBase'] = {'cidr': base_cidr}
        # current_base_address = first_network_address_block

        subnet_config = self._get_subnet_config_w_cidr(network_config)
        subnet_config = self._subnet_configs = list(subnet_config)

        for index, subnet_config in enumerate(subnet_config):
            # subnet_type = subnet_config.get('type', 'private')
            # subnet_size = subnet_config.get('size', '22')
            subnet_name = subnet_config.get('name', 'subnet')
            subnet_az = subnet_config.get('AZ', '-1')
            subnet_cidr = subnet_config.get('cidr', 'ERROR')
            az_key = 'AZ{}'.format(subnet_az)

            # TODO: check for subnet collisions

            if az_key not in ret_val:
                ret_val[az_key] = dict()
            if subnet_name not in ret_val[az_key]:
                ret_val[az_key][subnet_name] = dict()
            ret_val[az_key][subnet_name] = subnet_cidr

        return self.add_mapping('networkAddresses', ret_val)

    def add_vpn_gateway(self,
                        vpn_conf):
        """
        Not surprisingly, adds a VPN gateway to the network created by this template.
        @param vpn_conf [dict] - collection of vpn-level configuration values.
        """
        if 'vpn_name' in vpn_conf:
            vpn_name = vpn_conf.get('vpn_name')
        else:
            vpn_name = self.__class__.__name__ + 'Gateway'

        gateway = self.add_resource(
            ec2.VPNGateway(
                'vpnGateway',
                Type=vpn_conf.get('vpn_type', 'ipsec.1'),
                Tags=[ec2.Tag(key='Name', value=vpn_name)]))

        self.add_resource(
            ec2.VPCGatewayAttachment(
                'vpnGatewayAttachment',
                VpcId=self.vpc_id,
                InternetGatewayId=self.igw,
                VpnGatewayId=gateway))



from pkg_resources import resource_string
from pkg_resources import resource_exists
import yaml
import json
import os
import six


def _test_filelike(parent, basename, validator):
    """
    Tests various file extension to find the requested file like resource
    :param parent: parent directory
    :param basename: portion of filename excluding the file extension
    :return: file path of the resource that exists, or None
    """
    (basename, provided_suffix) = os.path.splitext(basename)
    suffix_list = EXTENSIONS
    if provided_suffix == '' or provided_suffix not in suffix_list:
        suffix_list = [provided_suffix] + EXTENSIONS

    for extension in suffix_list:
        file_path = os.path.join(parent, basename + extension)
        if validator(file_path):
            return file_path

    return None


def test_resource(parent, basename, relative_to_module_name=__name__):
    resource_test = lambda file_path: resource_exists(relative_to_module_name, file_path)
    return _test_filelike(parent, basename, resource_test)


def test_file(parent, basename):
    file_test = lambda file_path: os.path.isfile(file_path)
    return _test_filelike(parent, basename, file_test)


def get_yaml_resource(resource_name, relative_to_module_name=__name__):
    """
    Get package resource as json
    """
    return yaml.load(get_resource(resource_name, relative_to_module_name))


def get_resource(resource_name, relative_to_module_name=__name__):
    """
    Retrieves resource embedded in the package (even if installed as a zipped archive).
    """
    file_path = test_resource('data', resource_name, relative_to_module_name)
    file_content = resource_string(relative_to_module_name, file_path)
    return file_content


EXTENSIONS = ['.json', '.yaml', '.yml']

DEFAULT_CONFIG_FILENAME = 'config'
FACTORY_DEFAULT_CONFIG = get_yaml_resource(DEFAULT_CONFIG_FILENAME)


DEFAULT_AMI_CACHE_FILENAME = 'ami_cache'
FACTORY_DEFAULT_AMI_CACHE = get_yaml_resource(DEFAULT_AMI_CACHE_FILENAME)


CONFIG_REQUIREMENTS_FILENAME = 'config_schema'
CONFIG_REQUIREMENTS = get_yaml_resource(CONFIG_REQUIREMENTS_FILENAME)


COMMON_STRINGS_FILENAME = 'common_strings'
COMMON_STRINGS = get_yaml_resource(COMMON_STRINGS_FILENAME)


def load_file(parent, basename):
    file_path = test_file(parent, basename)
    if not file_path:
        # Whenever possible, don't use bare Exceptions like below (it's an anti-pattern)
        # Read more about it: https://realpython.com/blog/python/the-most-diabolical-python-antipattern/
        raise Exception("%s does not exist. Try running the 'init' command to generate it.\n" % (basename + EXTENSIONS[0]))

    return load_yaml_file(file_path)


def load_yaml_file(file_path):

    if not os.path.isfile(file_path):
        # Whenever possible, don't use bare Exceptions like below (it's an anti-pattern)
        # Read more about it: https://realpython.com/blog/python/the-most-diabolical-python-antipattern/
        raise Exception('{} does not exist'.format(file_path))

    with open(file_path, 'r') as f:
        try:
            content = f.read()
            parsed_content = yaml.load(content)
        except ValueError:
            print('%s could not be parsed' % file_path)
            raise

    return parsed_content


def load_json_file(file_path):

    if not os.path.isfile(file_path):
        # Whenever possible, don't use bare Exceptions like below (it's an anti-pattern)
        # Read more about it: https://realpython.com/blog/python/the-most-diabolical-python-antipattern/
        raise Exception('{} does not exist'.format(file_path))

    with open(file_path, 'r') as f:
        try:
            content = f.read()
            parsed_content = json.loads(content)
        except ValueError:
            print('%s could not be parsed' % file_path)
            raise

    return parsed_content


def get_str(key, default=None):
    return COMMON_STRINGS.get(key, default)


def get_type(typename):
    """
    Convert typename to type object
    :param typename: String name of type
    :return: __builtin__ type instance
    """
    types = {
        'bool': bool,
        'int': int,
        'float': float,
        # avoid all the python unicode weirdness by making all the strings basestrings
        'str': six.string_types,
        'basestring': six.string_types,
        'list': list
    }
    return types.get(typename, None)



from environmentbase.networkbase import NetworkBase
from environmentbase.template import Template
from environmentbase.environmentbase import EnvConfig
from troposphere import ec2


class MyRootTemplate(NetworkBase):
    """
    Class creates a VPC and common network components for the environment
    """
    def create_hook(self):
        self.add_child_template(MyChildTemplate('ChildTemplate'))


class MyChildTemplate(Template):
    """
    Class creates a VPC and common network components for the environment
    """

    # Called from add_child_template() after some common parameters are attached to this instance, see docs for details
    def build_hook(self):
        self.add_resource(ec2.Instance("ec2instance", InstanceType="m3.medium", ImageId="ami-e7527ed7"))

    # When no config.json file exists a new one is created using the 'factory default' file.  This function
    # augments the factory default before it is written to file with the config values required
    @staticmethod
    def get_factory_defaults():
        return {'my_child_template': {'favorite_color': 'blue'}}

    # When the user request to 'create' a new template the config.json file is read in. This file is checked to
    # ensure all required values are present. Because MyChildTemplate has additional requirements beyond that of
    # EnvironmentBase this function is used to add additional validation checks.
    @staticmethod
    def get_config_schema():
        return {'my_child_template': {'favorite_color': 'str'}}


if __name__ == '__main__':

    # EnvConfig holds references to handler classes used to extend certain functionality
    # of EnvironmentBase. The config_handlers list takes any class that implements
    # get_factory_defaults() and get_config_schema().
    env_config = EnvConfig(config_handlers=[MyChildTemplate])

    MyRootTemplate(env_config=env_config)



from unittest2 import TestCase, main
import mock
from mock import patch
import os
import shutil
import yaml
import json
import sys
import copy
from tempfile import mkdtemp
from environmentbase import cli, resources as res, environmentbase as eb
from environmentbase import networkbase
import environmentbase.patterns.ha_nat
from troposphere import ec2


class EnvironmentBaseTestCase(TestCase):

    def setUp(self):
        # Change to a temp dir so auto generated file don't clutter the os
        self.temp_dir = mkdtemp()
        os.chdir(self.temp_dir)

    def tearDown(self):
        # Delete any files left in the temp dir
        shutil.rmtree(self.temp_dir)
        assert not os.path.isdir(self.temp_dir)

    def fake_cli(self, extra_args):
        args = ['environmentbase']
        args.extend(extra_args)

        with patch.object(sys, 'argv', args):
            my_cli = cli.CLI(quiet=True)
            my_cli.process_request = mock.MagicMock()

        return my_cli

    def _create_dummy_config(self, env_base=None):
        dummy_string = 'dummy'
        dummy_bool = False
        dummy_int = 3
        dummy_list = ['A', 'B', 'C']

        config_requirements = copy.deepcopy(res.CONFIG_REQUIREMENTS)

        if env_base:
            for handler in env_base.config_handlers:
                config_requirements.update(handler.get_config_schema())

        config = {}
        for (section, keys) in config_requirements.iteritems():
            config[section] = {}
            for (key, key_type) in keys.iteritems():
                if key_type == basestring.__name__ or key_type == str.__name__:
                    config[section][key] = dummy_string
                elif key_type == bool.__name__:
                    config[section][key] = dummy_bool
                elif key_type == int.__name__:
                    config[section][key] = dummy_int
                elif key_type == list.__name__:
                    config[section][key] = dummy_list

        config['boto']['region_name'] = config['global']['valid_regions'][0]
        return config

    def _create_local_file(self, name, content):
        f = open(os.path.join(self.temp_dir, name), 'a')
        f.write(content)
        f.flush()
        return f

    def test_constructor(self):
        """Make sure EnvironmentBase passes control to view to process user requests"""
        fake_cli = self.fake_cli(['init'])
        env_base = eb.EnvironmentBase(fake_cli)

        # Check that EnvironmentBase started the CLI
        fake_cli.process_request.assert_called_once_with(env_base)

    def test_alternate_view(self):
        """ More of an example of how to use your own custom view than a test """
        actions_called = {'init': 0, 'deploy': 0, 'create': 0, 'delete': 0}

        class MyView(object):

            def __init__(self):
                super(MyView, self).__init__()
                # Start an api, a web server or a rich client UI for example
                # Record user request(s), the controller will then call process_request()
                # so the can relay user requests to the appropriate controller action
                self.user_actions = ['init', 'create', 'deploy', 'delete']
                self.user_config_changes = {'output_filename': 'output.txt'}

            def update_config(self, config):
                # Update any config properties you need to
                # config['global']['print_debug'] = self.user_config_changes['debug']
                config['global']['output'] = self.user_config_changes['output_filename']

            def process_request(self, controller):

                for action in self.user_actions:
                    # if action == 'create':
                    #     controller.create_action()
                    # elif action == 'deploy':
                    #     controller.deploy_action()

                    actions_called[action] += 1

        eb.EnvironmentBase(MyView())

        self.assertEqual(actions_called['init'], 1)
        self.assertEqual(actions_called['create'], 1)
        self.assertEqual(actions_called['deploy'], 1)
        self.assertEqual(actions_called['delete'], 1)

    def test_config_yaml(self):
        """ Make sure load_config can load yaml files."""
        with open("config.yaml", 'w') as f:
            f.write(yaml.dump(res.FACTORY_DEFAULT_CONFIG, default_flow_style=False))
            f.flush()

        fake_cli = self.fake_cli(['create', '--config-file', 'config.yaml'])
        base = eb.EnvironmentBase(fake_cli)
        base.load_config()

        self.assertEqual(base.config['global']['environment_name'], 'environmentbase')



    # def test_config_override(self):
    #     """ Make sure local config files overrides default values."""
    #
    #     # We don't care about the AMI cache for this test,
    #     # but the file has to exist and to contain valid json
    #     self._create_local_file(res.DEFAULT_AMI_CACHE_FILENAME + res.EXTENSIONS[0], '{}')
    #
    #     # Create a local config file and verify that it overrides the factory default
    #     config = self._create_dummy_config()
    #
    #     # Change one of the values
    #     original_value = config['global']['environment_name']
    #     config['global']['environment_name'] = original_value + 'dummy'
    #
    #     with open(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0], 'w') as f:
    #         f.write(yaml.dump(config))
    #         f.flush()
    #
    #     fake_cli = self.fake_cli(['create'])
    #     base = eb.EnvironmentBase(fake_cli)
    #     base.load_config()
    #
    #     self.assertNotEqual(base.config['global']['environment_name'], original_value)
    #
    #     # 4) Validate local config with non-default name
    #     config_filename = 'not_default_name'
    #
    #     # existence check
    #     with self.assertRaises(Exception):
    #         base = eb.EnvironmentBase(self.fake_cli(['create', '--config-file', config_filename]))
    #         base.load_config()
    #
    #     # remove config.json and create the alternate config file
    #     os.remove(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0])
    #     self.assertFalse(os.path.isfile(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0]))
    #
    #     with open(config_filename, 'w') as f:
    #         f.write(yaml.dump(config))
    #         f.flush()
    #         base = eb.EnvironmentBase(self.fake_cli(['create', '--config-file', config_filename]))
    #         base.load_config()
    #
    #     self.assertNotEqual(base.config['global']['environment_name'], original_value)
    #
    # def test_config_validation(self):
    #     """
    #     environmentbase.TEMPLATE_REQUIREMENTS defines the required sections and keys for a valid input config file
    #     This test ensures that EnvironmentBase._validate_config() enforces the TEMPLATE_REQUIREMENTS contract
    #     """
    #     cntrl = eb.EnvironmentBase(self.fake_cli(['create']))
    #
    #     valid_config = self._create_dummy_config()
    #     cntrl._validate_config(valid_config)
    #
    #     # Find a section with at least one required key
    #     section = ''
    #     keys = {}
    #     while True:
    #         (section, keys) = valid_config.items()[0]
    #         if len(keys) > 0:
    #             break
    #     assert len(keys) > 0
    #
    #     # Check type error
    #     with self.assertRaises(eb.ValidationError):
    #         invalid_config = copy.deepcopy(valid_config)
    #         invalid_config['global']['print_debug'] = "dfhkjdshf"
    #         cntrl._validate_config(invalid_config)
    #
    #     # Check missing key validation
    #     (key, value) = keys.items()[0]
    #     del valid_config[section][key]
    #
    #     with self.assertRaises(eb.ValidationError):
    #         cntrl._validate_config(valid_config)
    #
    #     # Check missing section validation
    #     del valid_config[section]
    #
    #     with self.assertRaises(eb.ValidationError):
    #         cntrl._validate_config(valid_config)
    #
    #     # Check wildcard sections
    #     extra_reqs = {'*-db': {'host': 'str', 'port': 'int'}}
    #     extra_reqs.update(res.CONFIG_REQUIREMENTS)
    #
    #     valid_config.update({
    #         'my-db': {'host': 'localhost', 'port': 3306},
    #         'my-other-db': {'host': 'localhost', 'port': 3306}
    #     })
    #
    #     # Check deep nested sections
    #     extra_reqs = {
    #         'lets': {
    #             'go': {
    #                 'deeper': {
    #                     'key': 'str'
    #                 }}}}
    #     extra_reqs.update(res.CONFIG_REQUIREMENTS)
    #
    #     valid_config.update({
    #         'lets': {
    #             'go': {
    #                 'deeper': {
    #                     'key': 'super_secret_value'
    #                 }}}})

    # def test_extending_config(self):
    #
    #     # Typically this would subclass eb.Template
    #     class MyConfigHandler(object):
    #         @staticmethod
    #         def get_factory_defaults():
    #             return {'new_section': {'new_key': 'value'}}
    #
    #         @staticmethod
    #         def get_config_schema():
    #             return {'new_section': {'new_key': 'str'}}
    #
    #     class MyEnvBase(eb.EnvironmentBase):
    #         pass
    #
    #     view = self.fake_cli(['init'])
    #     env_config=eb.EnvConfig(config_handlers=[MyConfigHandler])
    #     controller = MyEnvBase(
    #         view=view,
    #         env_config=env_config
    #     )
    #     controller.init_action()
    #     controller.load_config()
    #
    #     # Make sure the runtime config and the file saved to disk have the new parameter
    #     self.assertEquals(controller.config['new_section']['new_key'], 'value')
    #
    #     with open(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0], 'r') as f:
    #         external_config = yaml.load(f)
    #         self.assertEquals(external_config['new_section']['new_key'], 'value')
    #
    #     # Check extended validation
    #     # recreate config file without 'new_section' and make sure it fails validation
    #     os.remove(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0])
    #     dummy_config = self._create_dummy_config()
    #     self._create_local_file(res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0], json.dumps(dummy_config, indent=4))
    #
    #     with self.assertRaises(eb.ValidationError):
    #         base = MyEnvBase(view=view, env_config=env_config)
    #         base.load_config()


    def test_generate_config(self):
        """ Verify cli flags update config object """

        # Verify that debug and output are set to the factory default
        base = eb.EnvironmentBase(self.fake_cli(['init']))
        base.init_action()
        base.load_config()
        self.assertEqual(base.config['global']['print_debug'],
                         res.FACTORY_DEFAULT_CONFIG['global']['print_debug'])
        self.assertEqual(base.config['global']['environment_name'],
                         res.FACTORY_DEFAULT_CONFIG['global']['environment_name'])

    def test_template_file_flag(self):
        # verify that the --template-file flag changes the config value
        dummy_value = 'dummy'
        base = eb.EnvironmentBase(self.fake_cli(['create', '--template-file', dummy_value]))
        base.init_action()
        base.load_config()
        self.assertEqual(base.config['global']['environment_name'], dummy_value)

    def test_config_file_flag(self):
        dummy_value = 'dummy'
        base = eb.EnvironmentBase(self.fake_cli(['create', '--config-file', dummy_value]))
        base.init_action()
        self.assertTrue(os.path.isfile(dummy_value))

    def test_factory_default(self):
        with self.assertRaises(Exception):
            base = eb.EnvironmentBase(self.fake_cli(['init']))
            base.load_config()



        # Create refs to files that should be created and make sure they don't already exists
        config_file = os.path.join(self.temp_dir, res.DEFAULT_CONFIG_FILENAME + res.EXTENSIONS[0])
        ami_cache_file = os.path.join(self.temp_dir, res.DEFAULT_AMI_CACHE_FILENAME + res.EXTENSIONS[0])
        self.assertFalse(os.path.isfile(config_file))
        self.assertFalse(os.path.isfile(ami_cache_file))

        # Verify that create_missing_files works as intended
        base = eb.EnvironmentBase(self.fake_cli(['init']))
        base.init_action()
        self.assertTrue(os.path.isfile(config_file))
        # TODO: After ami_cache is updated change 'create_missing_files' to be singular
        # self.assertTrue(os.path.isfile(ami_cache_file))

        # Verify that the previously created files are loaded up correctly
        eb.EnvironmentBase(self.fake_cli(['create']))


    # The following two tests use a create_action, which currently doesn't test correctly

    # def test_controller_subclass(self):
    #     """ Example of out to subclass the Controller to provide additional resources """
    #     class MyController(eb.EnvironmentBase):
    #         def __init__(self, view):
    #             # Run parent initializer
    #             eb.EnvironmentBase.__init__(self, view)

    #         # Add some stuff
    #         def create_hook(self):
    #             res = ec2.Instance("ec2instance", InstanceType="m3.medium", ImageId="ami-951945d0")
    #             self.template.add_resource(res)


    #     # Initialize the the controller with faked 'create' CLI parameter
    #     with patch.object(sys, 'argv', ['environmentbase', 'init']):
    #         ctrlr = MyController(cli.CLI(quiet=True))
    #         ctrlr.load_config()
    #         ctrlr.create_action()

    #         # Load the generated output template
    #         template_path = os.path.join(ctrlr._ensure_template_dir_exists(), ctrlr.config['global']['environment_name'] + '.template')

    #         with open(template_path, 'r') as f:
    #             template = yaml.load(f)

    #         # Verify that the ec2 instance is in the output
    #         self.assertTrue('ec2instance' in template['Resources'])

    #         # print json.dumps(template, indent=4)

    # def test_nat_role_customization(self):
    #     """ Example of out to subclass the Controller to provide additional resources """
    #     class MyNat(environmentbase.patterns.ha_nat.HaNat):
    #         def get_extra_policy_statements(self):
    #           return [{
    #               "Effect": "Allow",
    #               "Action": ["DummyAction"],
    #               "Resource": "*"
    #           }]

    #     class MyController(networkbase.NetworkBase):

    #         def create_nat(self, index, nat_instance_type, enable_ntp, name, extra_user_data=None):
    #             return MyNat(index, nat_instance_type, enable_ntp, name, extra_user_data)

    #     # Initialize the the controller with faked 'create' CLI parameter

    #     ctrlr = MyController((self.fake_cli(['init'])))
    #     ctrlr.init_action()
    #     ctrlr.load_config()
    #     ctrlr.create_action()

    #     # Load the generated output template
    #     template_path = os.path.join(ctrlr._ensure_template_dir_exists(), ctrlr.config['global']['environment_name'] + '.template')

    #     with open(template_path, 'r') as f:
    #         template = yaml.load(f)

    #     # Verify that the ec2 instance is in the output
    #     self.assertIn('Nat0Role', template['Resources'])
    #     self.assertIn('Properties', template['Resources']['Nat0Role'])
    #     self.assertIn('Policies', template['Resources']['Nat0Role']['Properties'])
    #     self.assertEqual(len(template['Resources']['Nat0Role']['Properties']['Policies']), 1)
    #     policy = template['Resources']['Nat0Role']['Properties']['Policies'][0];
    #     self.assertIn('PolicyDocument', policy)
    #     self.assertIn('Statement', policy['PolicyDocument'])
    #     self.assertEqual(len(policy['PolicyDocument']['Statement']), 2)
    #     self.assertEqual(policy['PolicyDocument']['Statement'][1]['Action'], ['DummyAction'])


    # Cloudformation doesn't currently support a dry run, so this test would create a live stack
    # def test_deploy(self):
    #     with patch.object(sys, 'argv', [
    #         'environmentbase',
    #         'deploy',
    #         '--debug',
    #         '--template_file', os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources/amzn_linux_ec2.json')]):
    #         env_base = eb.EnvironmentBase()


if __name__ == '__main__':
    main()



__author__ = 'Eric Price'

from urllib2 import urlopen
from lxml.html import fromstring
import json

"""
Yes, it's a screen scraper script for assembling the instance type to arch map.
The thought is that updates to the web page will be minor ... maybe

You will need to have the following two dependencies to your environment: lxml, cssselect
"""

SELECT_TABLES = "div.informaltable"
SELECT_TABLE_DISCRIMINATOR = "table > thead > tr > th:nth-child(2)"

INSTANCE_TYPE_DELIMITER = ' | '


def get_page(url):
    html = urlopen(url).read()
    dom = fromstring(html)
    dom.make_links_absolute(url)
    return dom


def find_table(tbls, contains_text):
    return next(tbl for tbl in tbls if contains_text in tbl.cssselect(SELECT_TABLE_DISCRIMINATOR)[0].text)


def scrape_table_cells(tbl):
    tbl_cells = []
    for row in tbl.cssselect("tbody > tr"):
        content = row.cssselect('td:nth-child(2)')[0].text_content()
        tbl_cells.append(content)

    return tbl_cells


def get_instance_types(tbl):
    tbl_cells = scrape_table_cells(tbl)
    all_rows = INSTANCE_TYPE_DELIMITER.join(tbl_cells)
    instance_types_list = all_rows.split(INSTANCE_TYPE_DELIMITER)
    return instance_types_list


def build_type_to_arch_map(tbl, arch_type, map):
    for it in get_instance_types(tbl):
        map[it] = {'Arch': arch_type}


if __name__ == '__main__':
    dom = get_page("http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-types.html")

    tables = dom.cssselect(SELECT_TABLES)

    hvm_tbl = find_table(tables, "Current Generation")
    pv_tbl = find_table(tables, "Previous Generation")

    instance_type_to_arch_map = {}

    build_type_to_arch_map(hvm_tbl, 'HVM64', instance_type_to_arch_map)
    build_type_to_arch_map(pv_tbl, 'PV64', instance_type_to_arch_map)

    # Special case for 2x GPU instance with nvidia drivers
    instance_type_to_arch_map['g2.2xlarge']['Arch'] = 'HVMG2'

    json_str = json.dumps(instance_type_to_arch_map, indent=4, separators=(',', ': '), sort_keys=True)

    # Someone please explain to me why I have to do this repeatedly for all matches to be replaced!!
    import re
    for _ in range(7):
        json_str = re.sub(r": \{\n\s*([^\n]+)\n\s*\}", r": { \1 }", json_str, re.MULTILINE)
    print(json_str)



from troposphere import Output, Ref, Join, Parameter, Base64, GetAtt, FindInMap, Retain, Tags
from troposphere import iam, ec2, autoscaling, route53 as r53, s3, logs, cloudwatch
from awacs import logs as awacs_logs, aws
from awacs.helpers.trust import make_simple_assume_statement
import troposphere as t
import troposphere.elasticloadbalancing as elb
import troposphere.cloudformation as cf
import hashlib
import json
import os
from datetime import datetime
from . import resources as res
from . import utility

from toolz.dicttoolz import merge


DEFAULT_TO_MIN_SIZE = object()


class Template(t.Template):
    """
    Custom wrapper for Troposphere Template object which handles S3 uploads and a specific
    workflow around hashing the template to allow for a validation mechanism of a template's
    consistency since it was generated.
    """

    # Class variable for S3 destination, set once by controller for use across all templates w/in an environment
    s3_path_prefix = ''

    # S3 bucket name used to store the templates
    template_bucket_default = ''
    template_bucket_param = 'TemplateBucket'

    # Timeout period after which to fail if a child stack has not reached a COMPLETE state
    stack_timeout = '60'

    def __init__(self, template_name):
        """
        Init method for environmentbase.Template class
        @param template_name [string] - name of this template, used to identify this template when uploading, deploying, etc.
        """
        t.Template.__init__(self)
        self.name = template_name
        self.AWSTemplateFormatVersion = ''

        self._vpc_cidr = None
        self._vpc_id = None
        self._common_security_group = None
        self._utility_bucket = None
        self._child_templates = []
        self._child_template_references = []
        self.manual_parameter_bindings = {}

        self._subnets = {}

    def _ref_maybe(self, item):
        """
        Wraps provided item in a troposphere.Ref() if the type makes sense to ref in cloudformation.
        This allows attributes to be saved w/o needing to ref() them before (or after).
        Note: Dicts and Lists are recursively processed for 'ref'able values
        """
        # Wrap input if type is:
        # - AWSDeclaration --> Parameters & Outputs
        # - AWSObject --> Resources
        # - items in a list or valued of a hash
        if isinstance(item, (t.AWSDeclaration, t.AWSObject)):
            return Ref(item)

        elif isinstance(item, list):
            items = []
            for i in item:
                items.append(self._ref_maybe(i))
            return items

        elif isinstance(item, dict):
            items = {}
            for (k, v) in item.iteritems():
                items.update({k: self._ref_maybe(v)})
            return items

        else:
            return item

    @property
    def vpc_cidr(self):
        return self._ref_maybe(self._vpc_cidr)

    @property
    def vpc_id(self):
        return self._ref_maybe(self._vpc_id)

    @property
    def common_security_group(self):
        return self._ref_maybe(self._common_security_group)

    @property
    def utility_bucket(self):
        return self._ref_maybe(self._utility_bucket)

    @property
    def ec2_key(self):
        return self._ref_maybe(self._ec2_key)

    @property
    def subnets(self):
        return self._ref_maybe(self._subnets)

    def __get_template_hash(self):
        """
        Private method holds process for hashing this template for future validation.
        """
        m = hashlib.sha256()
        m.update(self.__validation_formatter())
        return m.hexdigest()

    def merge(self, other_template):
        """
        Experimental merge function
        1. This passes all the initialized attributes to the other template
        2. Calls the other template's build_hook()
        3. Copies the generated troposphere attributes back into this template

        NOTE: This function is currently used successfully in networkbase for merging the HaNat pattern
              into the root template. It has not been thoroughly tested in more complex scenarios
        """
        other_template.copy_attributes_from(self)

        other_template.build_hook()

        self.metadata.update(other_template.metadata)
        self.conditions.update(other_template.conditions)
        self.mappings.update(other_template.mappings)
        self.outputs.update(other_template.outputs)
        self.parameters.update(other_template.parameters)
        self.resources.update(other_template.resources)

    def copy_attributes_from(self, other_template):
        """
        Copies all attributes from the other template into this one
        These typically get initialized for a template when add_child_template is called
        from the controller, but that never happens when merging two templates
        """
        self._vpc_cidr = other_template._vpc_cidr
        self._vpc_id = other_template._vpc_id
        self._common_security_group = other_template._common_security_group
        self._utility_bucket = other_template._utility_bucket

        self._subnets = other_template.subnets.copy()

        self.parameters = other_template.parameters.copy()
        self.mappings = other_template.mappings.copy()
        self.metadata = other_template.metadata.copy()
        self.conditions = other_template.conditions.copy()
        self.outputs = other_template.outputs.copy()
        self.resources = other_template.resources.copy()

    def build_hook(self):
        """
        Provides template subclasses a place to assemble resources with access to common parameters and mappings.
        Executed by add_child_template() after add_common_params_to_child_template() and load_ami_cache()
        """
        pass

    @staticmethod
    def get_config_schema():
        """
        This method is provided for subclasses to update config requirements with additional required keys and their types.
        If you define both this and the get_factory_defaults() functions in your template and register the template with
        the controller's config handlers, these values will be included when you run the `init` command

        The format is a dictionary with key values being one of bool/int/float/str/list.
        Example (yes comments are allowed):
        {
            "template": {
                // Name of json file containing mapping labels to AMI ids
                "ami_map_file": "basestring",
                "mock_upload": "bool",
            }
        }
        :return: dict of config settings to be merged into base config, match existing keys to replace.
        """
        return {}

    @staticmethod
    def get_factory_defaults():
        """
        This method is provided for subclasses to update factory default config file with additional sections.
        If you define both this and the get_config_schema() functions in your template and register the template with
        the controller's config handlers, these values will be included when you run the `init` command

        The format is basic json (with comment support).
        {
            "template": {
                // Name of json file containing mapping labels to AMI ids
                "ami_map_file": "ami_cache.json",
                "mock_upload": false,
            }
        }
        :return: dict of config settings to be merged into base config, match existing keys to replace.
        """
        return {}

    def to_template_json(self):
        """
        Process all child templates recursively and render this template as json with a timestamp identifying
        when it was generated along with a SHA256 hash representing the template for validation purposes
        """
        self.process_child_templates()

        # strip existing values
        for output_key in ['dateGenerated', 'templateValidationHash']:
            if output_key in self.outputs:
                self.outputs.pop(output_key)

        # generate the template validation hash
        if self.include_templateValidationHash_output and 'templateValidationHash' not in self.outputs:
            self.add_output(Output(
                'templateValidationHash',
                Value=self.__get_template_hash(),
                Description='''Hash of this template that can be used as a simple means of validating whether a template has been changed
                since it was generated.'''))

        # set the date that this template was generated
        if self.include_dateGenerated_output and 'dateGenerated' not in self.outputs:
            self.add_output(Output(
                'dateGenerated',
                Value=str(datetime.utcnow()),
                Description='UTC datetime representation of when this template was generated'))

        return self.to_json()

    def __validation_formatter(self):
        """
        Validation formatter helps to ensure consistent formatting for hash validation workflow
        """
        return json.dumps(json.loads(self.to_json()), separators=(',', ':'))

    def add_parameter_idempotent(self, troposphere_parameter):
        """
        Idempotent add (add only if not exists) for parameters within the template
        @param [Troposphere.Parameter] Troposphere Parameter to add to this template
        """
        if troposphere_parameter.title not in self.parameters:
            return self.add_parameter(troposphere_parameter)
        else:
            return None

    def add_instance_profile(self, layer_name, iam_policies, path_prefix):
        """
        Helper function to add role and instance profile resources to this template
        using the provided iam_policies. The instance_profile will be created at:
        '/<path_prefix>/<layer_name>/'
        """
        iam_role_obj = iam.Role(
            layer_name + 'IAMRole',
            AssumeRolePolicyDocument={
                'Statement': [{
                    'Effect': 'Allow',
                    'Principal': {'Service': ['ec2.amazonaws.com']},
                    'Action': ['sts:AssumeRole']
                }]
            },
            Path=Join('', ['/' + path_prefix + '/', layer_name, '/']))

        if iam_policies is not None:
            iam_role_obj.Policies = iam_policies

        iam_role = self.add_resource(iam_role_obj)

        return self.add_resource(
            iam.InstanceProfile(
                layer_name + 'InstancePolicy',
                Path='/' + path_prefix + '/',
                Roles=[Ref(iam_role)])
        )

    def add_common_parameters_from_parent(self, parent):
        ec2_key = parent._ec2_key.Default
        parent_subnets = parent._subnets if not self._subnets else {}

        # Merge RegionMap w/ parent's because there's only 1 ami_cache
        # TODO: allow different ami_mappings per child template
        if 'RegionMap' in parent.mappings:
            if 'RegionMap' in self.mappings:
                region_map = dict(self._merge_region_map(self.mappings['RegionMap'], parent.mappings['RegionMap']))
            else:
                region_map = parent.mappings['RegionMap']

            self.mappings['RegionMap'] = region_map

        self.add_common_parameters(ec2_key, parent_subnets)

    def _merge_region_map(self, map1, map2):
        for key in set(map1.keys() + map2.keys()):
            yield (key, merge(map1[key], map2[key]))

    def add_common_parameters(self, ec2_key, parent_subnets):
        """
        Adds the common set of parameters that are available to every child template
        The values are automatically matched from the root template
            vpcCidr,
            vpcId,
            commonSecurityGroup,
            utilityBucket,
            each subnet: [public|private]Subnet[0-9]
        """
        self._vpc_cidr = self.add_parameter(Parameter(
            'vpcCidr',
            Description='CIDR of the VPC network',
            Type='String',
            AllowedPattern=res.get_str('cidr_regex'),
            ConstraintDescription=res.get_str('cidr_regex_message')))

        self._vpc_id = self.add_parameter(Parameter(
            'vpcId',
            Description='ID of the VPC network',
            Type='String'))

        self._common_security_group = self.add_parameter(Parameter(
            'commonSecurityGroup',
            Description='Security Group ID of the common security group for this environment',
            Type='String'))

        self._utility_bucket = self.add_parameter(Parameter(
            'utilityBucket',
            Description='Name of the S3 bucket used for infrastructure utility',
            Type='String'))

        self._ec2_key = self.add_parameter(
            Parameter(
                'ec2Key',
                Type='String',
                Default=ec2_key,
                Description='Name of an existing EC2 KeyPair to enable SSH access to the instances',
                AllowedPattern=res.get_str('ec2_key'),
                MinLength=1,
                MaxLength=255,
                ConstraintDescription=res.get_str('ec2_key_message')
            ))

        self.template_bucket_param = self.add_parameter(Parameter(
            Template.template_bucket_param,
            Type='String',
            Default=Template.template_bucket_default,
            Description='S3 Bucket where the templates will be stored.',
            AlloGreaterThanThresholdwedPattern=res.get_str('ascii_only'),
            MinLength=1,
            MaxLength=255,
            ConstraintDescription=res.get_str('ascii_only_message')
        ))

        for subnet_type in parent_subnets:
            if subnet_type not in self._subnets:
                self._subnets[subnet_type] = {}

            for subnet_layer in parent_subnets[subnet_type]:
                if subnet_layer not in self._subnets[subnet_type]:
                    self._subnets[subnet_type][subnet_layer] = []

                for subnet in parent_subnets[subnet_type][subnet_layer]:
                    if isinstance(subnet, Parameter):
                        subnet_name = subnet.title
                    else:
                        subnet_name = subnet.data['Ref']
                    self._subnets[subnet_type][subnet_layer].append(self.add_parameter(Parameter(
                        subnet_name,
                        Description=subnet_name,
                        Type='String')))

    @staticmethod
    def construct_user_data(env_vars={}, user_data=''):
        """
        Wrapper method to encapsulate process of constructing userdata for a launch configuration
        @param env_vars [dict] A dictionary containining key value pairs to set as environment variables in the userdata
        @param user_data [string] Contents of the user data script as a string
        Returns user_data_payload [string[]] Userdata payload ready to be dropped into a launch configuration
        """
        # At least one of env_vars or user_data must exist
        if not (env_vars or user_data):
            return []

        # If the variable value is not a string, use the Join function
        # This handles Refs, Parameters, etc. which are evaluated at runtime
        variable_declarations = []
        for k, v in env_vars.iteritems():
            if isinstance(v, basestring):
                variable_declarations.append('%s=%s' % (k, v))
            else:
                variable_declarations.append(Join('=', [k, v]))

        return Template.build_bootstrap(
            bootstrap_files=[user_data],
            variable_declarations=variable_declarations
        )

    @staticmethod
    def build_bootstrap(bootstrap_files=None,
                        variable_declarations=None,
                        cleanup_commands=None,
                        prepend_line='#!/bin/bash'):
        """
        Method encapsulates process of building out the bootstrap given a set of variables and a bootstrap file to source from
        Returns base 64-wrapped, joined bootstrap to be applied to an instnace
        @param bootstrap_files [ string[] ] list of paths to the bash script(s) to read as the source for the bootstrap action to created
        @param variable_declaration [ list ] list of lines to add to the head of the file - used to inject bash variables into the script
        @param cleanup_commnds [ string[] ] list of lines to add at the end of the file - used for layer-specific details
        """
        if prepend_line != '':
            ret_val = [prepend_line]
        else:
            ret_val = []

        if variable_declarations is not None:
            for line in variable_declarations:
                ret_val.append(line)
        for file_name_or_content in bootstrap_files:
            for line in Template.get_file_contents(file_name_or_content):
                ret_val.append(line)
        if cleanup_commands is not None:
            for line in cleanup_commands:
                ret_val.append(line)
        return Base64(Join("\n", ret_val))

    @staticmethod
    def get_file_contents(file_name_or_content):
        """
        Method encpsulates reading a file into a list while removing newline characters.
        If file is not found the variable is interpreted as the file content itself.
        @param file_name_or_content [string] path to file to read or content itself
        """
        ret_val = []
        if not os.path.isfile(file_name_or_content):
            content = file_name_or_content.split('\n')
        else:
            with open(file_name_or_content) as f:
                content = f.readlines()

        for line in content:
            if not line.startswith('#~'):
                ret_val.append(line.replace("\n", ""))
        return ret_val

    def add_ami_mapping(self, json_data):
        """
        Method gets the ami cache from the file locally and adds a mapping for ami ids per region into the template
        This depends on populating ami_cache.json with the AMI ids that are output by the packer scripts per region
        @param ami_map_file [string] path representing where to find the AMI map to ingest into this template
        """
        for region in json_data:
            for key in json_data[region]:
                self.add_region_map_value(region, key, json_data[region][key])

    def add_region_map_value(self,
                             region,
                             key,
                             value):
        """
        Method adds a key value pair to the RegionMap mapping within this CloudFormation template
        @param region [string] AWS region name that the key value pair is associated with
        @param key [string] name of the key to store in the RegionMap mapping for the specified Region
        @param value [string] value portion of the key value pair related to the region specified
        """
        self.__init_region_map([region])
        if region not in self.mappings['RegionMap']:
            self.mappings['RegionMap'][region] = {}
        self.mappings['RegionMap'][region][key] = value

    def __init_region_map(self,
                          region_list):
        """
        Internal helper method used to check to ensure mapping dictionaries are present
        @param region_list [list(str)] array of strings representing the names of the regions to validate and/or create within the
        RegionMap CloudFormation mapping
        """
        if 'RegionMap' not in self.mappings:
            self.mappings['RegionMap'] = {}
        for region_name in region_list:
            if region_name not in self.mappings['RegionMap']:
                self.mappings['RegionMap'][region_name] = {}

    def add_scaling_policy(
            self,
            metric_name,
            asg_name,
            adjustment_type="ChangeInCapacity",
            cooldown=1,
            scaling_adjustment=1):
        """
        Helper method to encapsulate process of adding a scaling policy to an autoscaling group in this template
        """
        policy = autoscaling.ScalingPolicy(
            metric_name + 'ScalingPolicy',
            AdjustmentType=adjustment_type,
            AutoScalingGroupName=Ref(asg_name),
            Cooldown=cooldown,
            ScalingAdjustment=str(scaling_adjustment))

        return self.add_resource(policy)

    def add_cloudwatch_alarm(
            self,
            layer_name,
            scaling_policy_name,
            asg_name,
            metric_name='CPUUtilization',
            evaluation_periods=1,
            statistic='Average',
            threshold=10,
            period=60,
            namespace='AWS/EC2',
            comparison_operator='GreaterThanThreshold'):
        """
        Helper method to encapsulate process of adding a cloudwatch alarm resource to a scaling policy
        """
        alarm = cloudwatch.Alarm(
            layer_name + 'Alarm',
            EvaluationPeriods=str(evaluation_periods),
            Statistic=statistic,
            Threshold=str(threshold),
            Period=str(period),
            AlarmActions=[Ref(scaling_policy_name)],
            Namespace=namespace,
            Dimensions=[cloudwatch.MetricDimension(Name="AutoScalingGroupName", Value=Ref(asg_name))],
            ComparisonOperator=comparison_operator,
            MetricName=metric_name)

        return self.add_resource(alarm)

    def add_asg(self,
                layer_name,
                instance_profile=None,
                instance_type='t2.micro',
                ami_name='amazonLinuxAmiId',
                ec2_key=None,
                user_data=None,
                target_groups=None,
                security_groups=None,
                min_size=1,
                max_size=1,
                desired_capacity=DEFAULT_TO_MIN_SIZE,
                root_volume_size=None,
                root_volume_type=None,
                include_ephemerals=True,
                number_ephemeral_vols=2,
                ebs_data_volumes=None,  # [{'size':'100', 'type':'gp2', 'delete_on_termination': True, 'iops': 4000, 'volume_type': 'io1'}],
                scaling_policies=None,  # [{'metric_name':'MyMetric', 'comparison_operator':'GTT', threshold':10, scaling_adjustment: 1}]
                custom_tags=None,
                load_balancer=None,
                health_check_type='EC2',
                health_check_grace_period=0,
                instance_monitoring=False,
                subnet_layer=None,
                associate_public_ip=None,
                launch_config_metadata=None,
                creation_policy=None,
                update_policy=None,
                depends_on=[]):
        """
        Wrapper method used to create an EC2 Launch Configuration and Auto Scaling group
        @param layer_name [string] friendly name of the set of instances being created - will be set as the name for instances deployed
        @param instance_profile [Troposphere.iam.InstanceProfile] IAM Instance Profile object to be applied to instances launched within
        this Auto Scaling group
        @param instance_type [Troposphere.Parameter | string] Reference to the AWS EC2 Instance Type to deploy.
        @param ami_name [string] Name of the AMI to deploy as defined within the RegionMap lookup for the deployed region
        @param ec2_key [Troposphere.Parameter | Troposphere.Ref(Troposphere.Parameter)] Input parameter used to gather the name of the EC2
        key to use to secure access to instances launched within this Auto Scaling group
        @param user_data [string[]] Array of strings (lines of bash script) to be set as the user data as a bootstrap script for instances
        launched within this Auto Scaling group
        @param security_groups [Troposphere.ec2.SecurityGroup[]] array of security groups to be applied to instances within this Auto
        Scaling group
        @param min_size [int|Parameter] value to set as the minimum number of instances for the Auto Scaling group
        @param max_size [int|Parameter] value to set as the maximum number of instances for the Auto Scaling group
        @param desired_capacity [int|Parameter] value to set as the maximum number of instances for the Auto Scaling group
        @param root_volume_size [int] size (in GiB) to assign to the root volume of the launched instance
        @param include_ephemerals [Boolean] indicates that ephemeral volumes should be included in the block device mapping of the Launch
        Configuration
        @param number_ephemeral_vols [int] number of ephemeral volumes to attach within the block device mapping Launch Configuration
        @param ebs_data_volumes [list] dictionary pair of size and type data properties in a list used to create ebs volume attachments
        @param scaling_policies [list] dictionaries describing scaling policies and their associated cloudwatch alarms
        @param custom_tags [Troposphere.autoscaling.Tag[]] Collection of Auto Scaling tags to be assigned to the Auto Scaling Group
        @param load_balancer [Troposphere.elasticloadbalancing.LoadBalancer] Object reference to an ELB to be assigned to this auto scaling
         group
        @param instance_monitoring [Boolean] indicates that detailed monitoring should be turned on for all instnaces launched within this
        Auto Scaling group
        @param subnet_layer [string] string indicating which subnet layer instances are being launched into
        """

        # Ensure that all the passed in parameters are Ref objects
        if ec2_key and type(ec2_key) != Ref:
            ec2_key = Ref(ec2_key)
        elif ec2_key is None:
            ec2_key = Ref(self.parameters['ec2Key'])

        if type(instance_type) != str:
            instance_type = Ref(instance_type)

        sg_list = []
        for sg in security_groups:
            if isinstance(sg, Ref):
                sg_list.append(sg)
            else:
                sg_list.append(Ref(sg))

        # If no instance profile was provided, create one with just the cloudformation read policy
        if not instance_profile:
            instance_profile = self.add_instance_profile(layer_name, [self.get_cfn_policy()], self.name)

        # If subnet_layer isn't passed in, try a private subnet if available, else a public subnet
        if not subnet_layer:
            if len(self._subnets.get('private')) > 0:
                subnet_layer = self._subnets['private'].keys()[0]
            else:
                subnet_layer = self._subnets['public'].keys()[0]

        subnet_type = self.get_subnet_type(subnet_layer)

        # If associate_public_ip is not passed in, set it based on the subnet_type
        if not associate_public_ip:
            associate_public_ip = True if subnet_type == 'public' else False

        launch_config_obj = autoscaling.LaunchConfiguration(
            layer_name + 'LaunchConfiguration',
            IamInstanceProfile=Ref(instance_profile),
            ImageId=FindInMap('RegionMap', Ref('AWS::Region'), ami_name),
            InstanceType=instance_type,
            SecurityGroups=sg_list,
            KeyName=ec2_key,
            AssociatePublicIpAddress=associate_public_ip,
            InstanceMonitoring=instance_monitoring)

        if launch_config_metadata:
            launch_config_obj.Metadata = launch_config_metadata

        if user_data:
            launch_config_obj.UserData = user_data

        block_devices = []
        if root_volume_type and root_volume_size:
            ebs_device = ec2.EBSBlockDevice(
                VolumeSize=root_volume_size)

            if root_volume_type:
                ebs_device.VolumeType = root_volume_type

            block_devices.append(ec2.BlockDeviceMapping(
                DeviceName='/dev/sda1',
                Ebs=ebs_device))

        device_names = ['/dev/sd%s' % c for c in 'bcdefghijklmnopqrstuvwxyz']

        if ebs_data_volumes is not None and len(ebs_data_volumes) > 0:
            for ebs_volume in ebs_data_volumes:
                # Respect names provided by AMI when available
                if 'name' in ebs_volume:
                    device_name = ebs_volume.get('name')
                    device_names.remove(device_name)
                else:
                    device_name = device_names.pop()

                ebs_block_device = ec2.EBSBlockDevice(
                    DeleteOnTermination=ebs_volume.get('delete_on_termination', True),
                    Encrypted=ebs_volume.get('encrypted', False),
                    VolumeSize=ebs_volume.get('size', '100'),
                    VolumeType=ebs_volume.get('type', 'gp2'))

                if 'iops' in ebs_volume:
                    ebs_block_device.Iops = int(ebs_volume.get('iops'))
                if 'snapshot_id' in ebs_volume:
                    ebs_block_device.SnapshotId = ebs_volume.get('snapshot_id')

                block_devices.append(ec2.BlockDeviceMapping(
                    DeviceName=device_name,
                    Ebs=ebs_block_device))

        if include_ephemerals and number_ephemeral_vols > 0:
            device_names.reverse()
            for x in range(0, number_ephemeral_vols):
                device_name = device_names.pop()
                block_devices.append(ec2.BlockDeviceMapping(
                    DeviceName=device_name,
                    VirtualName='ephemeral' + str(x)))

        if len(block_devices) > 0:
            launch_config_obj.BlockDeviceMappings = block_devices

        launch_config = self.add_resource(launch_config_obj)

        auto_scaling_obj = autoscaling.AutoScalingGroup(
            layer_name + 'AutoScalingGroup',
            LaunchConfigurationName=Ref(launch_config),
            MaxSize=max_size,
            MinSize=min_size,
            DesiredCapacity=min_size if desired_capacity == DEFAULT_TO_MIN_SIZE else desired_capacity,
            # DesiredCapacity=min_size if min_size <= max_size else max_size,  # consistent min algorithm
            VPCZoneIdentifier=self.subnets[subnet_type][subnet_layer.lower()],
            TerminationPolicies=['OldestLaunchConfiguration', 'ClosestToNextInstanceHour', 'Default'],
            DependsOn=depends_on,
            HealthCheckGracePeriod=health_check_grace_period,
            HealthCheckType=health_check_type,
            TargetGroupARNs=target_groups)

        lb_tmp = []

        if load_balancer is not None:
            try:
                if type(load_balancer) is dict:
                    for lb in load_balancer:
                        lb_tmp.append(Ref(load_balancer[lb]))
                elif type(load_balancer) is not Ref:
                    for lb in load_balancer:
                        lb_tmp.append(Ref(lb))
                else:
                    lb_tmp.append(load_balancer)
            except TypeError:
                lb_tmp.append(Ref(load_balancer))
        else:
            lb_tmp = None

        if lb_tmp is not None and len(lb_tmp) > 0:
            auto_scaling_obj.LoadBalancerNames = lb_tmp

        if creation_policy is not None:
            auto_scaling_obj.resource['CreationPolicy'] = creation_policy

        if update_policy is not None:
            auto_scaling_obj.resource['UpdatePolicy'] = update_policy

        if custom_tags is not None and len(custom_tags) > 0:
            final_custom_tags = []
            if type(custom_tags) == dict:
                for key, value in custom_tags.iteritems():
                    final_custom_tags.append(autoscaling.Tag(key, value, True))
            elif type(custom_tags) != list:
                final_custom_tags = [custom_tags]

            auto_scaling_obj.Tags = final_custom_tags
        else:
            auto_scaling_obj.Tags = []

        if scaling_policies is not None and len(scaling_policies) > 0:
            for scaling_policy in scaling_policies:
                policy_obj = self.add_scaling_policy(
                    scaling_policy.get('metric_name'),
                    auto_scaling_obj.name,
                    scaling_adjustment=scaling_policy.get('scaling_adjustment'),
                    adjustment_type=scaling_policy.get('adjustment_type', 'ChangeInCapacity'),
                    cooldown=scaling_policy.get('cooldown', 1))

                self.add_cloudwatch_alarm(
                    scaling_policy.get('metric_name'),
                    policy_obj.name,
                    auto_scaling_obj.name,
                    metric_name=scaling_policy.get('metric_name'),
                    threshold=scaling_policy.get('threshold'),
                    comparison_operator=scaling_policy.get('comparison_operator', 'GreaterThanThreshold'),
                    evaluation_periods=scaling_policy.get('evaluation_periods', 1),
                    statistic=scaling_policy.get('statistic', 'Average'),
                    period=scaling_policy.get('period', 60),
                    namespace=scaling_policy.get('namespace', 'AWS/EC2'))

        auto_scaling_obj.Tags.append(autoscaling.Tag('Name', self._autoscaling_name_tag_value(), True))
        return self.add_resource(auto_scaling_obj)

    def _autoscaling_name_tag_value(self):
        """
        Overridable method to assign a 'Name' tag with the returned value.
        """
        return self.name

    def add_elb(self,
                resource_name,
                listeners,
                utility_bucket=None,
                elb_custom_tags=None,
                instances=[],
                security_groups=[],
                depends_on=[],
                subnet_layer=None,
                scheme='internet-facing',
                health_check_protocol='TCP',
                health_check_port=None,
                health_check_path='',
                connection_draining_timeout=None,  # AWS default is 300 seconds
                cookie_expiration_period=None,
                idle_timeout=None):
        """
        Helper function creates an ELB and attaches it to your template
        Listeners should be a list of dictionaries, each containining:
            elb_port - The port of the incoming connection to the ELB
            elb_protocol (optional) - The protocol of the incoming connection to the ELB (i.e., 'HTTP', 'HTTPS', 'TCP', 'SSL')
            instance_port (optional) - The port of the incoming connection to the instances
            instance_protocol (optional) - The protocol of the incoming connection to the instances (i.e., 'HTTP', 'HTTPS', 'TCP', 'SSL')
            ssl_cert_name (optional) - The name of the SSL cert in IAM (Only required if the elb_protocol is 'HTTPS' or 'SSL')
        """
        if subnet_layer:
            subnet_type = self.get_subnet_type(subnet_layer)
        else:
            # If subnet layer is not passed in, determine based on the scheme
            # -- Pick a public subnet if it's internet-facing, else pick a private one
            subnet_type = 'public' if scheme == 'internet-facing' else 'private'
            subnet_layer = self._subnets[subnet_type].keys()[0]

        # Add optional parameters for LoadBalancer to this dictionary
        optional_elb_kwargs = {}

        # ConnectionDrainingPolicy
        if connection_draining_timeout is not None:
            connection_draining_policy = self.get_elb_connection_draining_policy(Timeout=connection_draining_timeout)
            optional_elb_kwargs["ConnectionDrainingPolicy"] = connection_draining_policy

        # LBCookieStickinessPolicy
        # TODO: Allow for a *list* of policies to be defined from input parameters
        http_stickiness_policy_names = []

        if cookie_expiration_period is not None:
            stickiness_policy_name = '%sElbStickinessPolicy' % resource_name
            stickiness_policy = self.get_elb_stickiness_policy(stickiness_policy_name, CookieExpirationPeriod=cookie_expiration_period)
            http_stickiness_policy_names = [stickiness_policy_name]
            optional_elb_kwargs["LBCookieStickinessPolicy"] = [stickiness_policy]

        elb_listeners = self.get_elb_listeners(listeners, http_stickiness_policy_names=http_stickiness_policy_names)

        elb_obj = elb.LoadBalancer(
            '%sElb' % resource_name,
            Subnets=self.subnets[subnet_type][subnet_layer],
            SecurityGroups=[Ref(sg) for sg in security_groups],
            CrossZone=True,
            Listeners=elb_listeners,
            Instances=instances,
            Scheme=scheme,
            DependsOn=depends_on,
            **optional_elb_kwargs
        )

        # If the health_check_port was not specified, use the instance port of any of the listeners
        # (elb_port is used if instance_port isn't set)
        if not health_check_port:
            health_check_port = [listener.get('instance_port')
                                 if listener.get('instance_port') else listener.get('elb_port') for listener in listeners][0]

        # Construct the ELB Health Check target based on the passed in health_check_protocol and health_check_port parameters
        health_check_protocol = health_check_protocol.upper()
        health_check_target = "%s:%s" % (health_check_protocol, health_check_port)

        # Add the health check path for HTTP/S targets (i.e., '/version', '/about', etc.)
        if health_check_protocol == 'HTTP' or health_check_protocol == 'HTTPS':
            # Ensure exactly one '/' at the beginning of the path string
            health_check_target += "/%s" % health_check_path.lstrip('/')

        # TODO: Parameterize this stuff
        elb_obj.HealthCheck = elb.HealthCheck(
            HealthyThreshold=3,
            UnhealthyThreshold=5,
            Interval=30,
            Target=health_check_target,
            Timeout=5
        )

        # If an S3 utility bucket was passed in, set up the ELB access log
        if utility_bucket is not None:
            elb_obj.AccessLoggingPolicy = elb.AccessLoggingPolicy(
                EmitInterval=5,
                Enabled=True,
                S3BucketPrefix=resource_name,
                S3BucketName=utility_bucket)

        # If the idle_timeout was passed in, create a ConnectionSettings object with the Idle Timeout
        if idle_timeout:
            elb_obj.ConnectionSettings = elb.ConnectionSettings(IdleTimeout=idle_timeout)

        # Add custom tags
        if elb_custom_tags:
            elb_obj.Tags = Tags(**elb_custom_tags)

        return self.add_resource(elb_obj)

    def get_elb_listeners(self, listeners, http_stickiness_policy_names):
        # Construct the listener objects based on the passed in listeners dictionary
        elb_listeners = []
        for listener in listeners:

            elb_port = listener.get('elb_port')
            elb_protocol = listener.get('elb_protocol').upper() if listener.get('elb_protocol') else 'TCP'

            # If back end parameters are not specified, use the values for the front end
            instance_port = listener.get('instance_port') if listener.get('instance_port') else elb_port
            instance_protocol = listener.get('instance_protocol').upper() if listener.get('instance_protocol') else elb_protocol

            elb_listener = elb.Listener(Protocol=elb_protocol,
                                        LoadBalancerPort=elb_port,
                                        InstanceProtocol=instance_protocol,
                                        InstancePort=instance_port)

            # SSL Cert must be included if using either SSL or HTTPS for elb_protocol
            ssl_cert_name = listener.get('ssl_cert_name')
            if ssl_cert_name:
                elb_listener.SSLCertificateId = Join("", [
                    "arn:aws:iam::",
                    {"Ref": "AWS::AccountId"},
                    ":server-certificate/",
                    ssl_cert_name])

            # Create the default session stickiness policy for HTTP or HTTPS listeners
            # TODO: Parameterize whether or not to include this policy
            if elb_protocol == 'HTTP' or elb_protocol == 'HTTPS':
                elb_listener.PolicyNames = http_stickiness_policy_names

            elb_listeners.append(elb_listener)
        return elb_listeners

    def get_elb_connection_draining_policy(self, Enabled=True, **kwargs):
        """Output like:
        {
           "Enabled" : Boolean,
           "Timeout" : Integer
        }
        Source: http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-elb-connectiondrainingpolicy.html
        """
        connection_draining_policy_kwargs = {"Enabled": Enabled, }
        connection_draining_policy_kwargs.update(kwargs)

        connection_draining_policy = elb.ConnectionDrainingPolicy(**connection_draining_policy_kwargs)
        return connection_draining_policy

    def get_elb_stickiness_policy(self, PolicyName, **kwargs):
        """Output like:
        {
           "PolicyName" : String
           "CookieExpirationPeriod" : String,
        }
        Source: http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-elb-connectiondrainingpolicy.html
        """
        stickiness_policy_kwargs = {"PolicyName": PolicyName, }
        stickiness_policy_kwargs.update(kwargs)

        stickiness_policy = elb.LBCookieStickinessPolicy(**stickiness_policy_kwargs)
        return stickiness_policy

    def create_reciprocal_sg(self,
                             source_group,
                             source_group_name,
                             destination_group,
                             destination_group_name,
                             from_port,
                             to_port=None,
                             ip_protocol='tcp'):
        """
        Helper method creates reciprocal ingress and egress rules given two existing security groups and a range of ports
        @param source_group [Troposphere.ec2.SecurityGroup] Object reference to the source security group
        @param source_group_name [string] friendly name of the source security group used for labels
        @param destination_group [Troposphere.ec2.SecurityGroup] Object reference to the destination security group
        @param destination_group_name [string] friendly name of the destination security group used for labels
        @param from_port [string] lower boundary of the port range to set for the secuirty group rules
        @param to_port [string] upper boundary of the port range to set for the security group rules
        @param ip_protocol [string] name of the IP protocol to set this rule for
        """
        if to_port is None:
            to_port = from_port
        if isinstance(from_port, unicode):
            from_port = from_port.encode('ascii', 'ignore')
        if isinstance(to_port, unicode):
            to_port = to_port.encode('ascii', 'ignore')
        if from_port == to_port:
            label_suffix = ip_protocol.capitalize() + str(from_port)
        else:
            label_suffix = ip_protocol.capitalize() + str(from_port) + 'To' + str(to_port)

        # A Ref cannot be created from an object that is already a GetAtt
        # and possibly some other CFN types, so expand this list if you discover another one
        CFN_TYPES = [GetAtt]
        if type(source_group) not in CFN_TYPES:
            source_group = Ref(source_group)
        if type(destination_group) not in CFN_TYPES:
            destination_group = Ref(destination_group)

        self.add_resource(ec2.SecurityGroupIngress(
            destination_group_name + 'Ingress' + source_group_name + label_suffix,
            SourceSecurityGroupId=source_group,
            GroupId=destination_group,
            FromPort=from_port,
            ToPort=to_port,
            IpProtocol=ip_protocol))

        self.add_resource(ec2.SecurityGroupEgress(
            source_group_name + 'Egress' + destination_group_name + label_suffix,
            DestinationSecurityGroupId=destination_group,
            GroupId=source_group,
            FromPort=from_port,
            ToPort=to_port,
            IpProtocol=ip_protocol))

    def get_cfn_policy(self):
        """
        Helper method returns the standard IAM policy to allow cloudformation read actions
        """
        return iam.Policy(
            PolicyName='cloudformationRead',
            PolicyDocument={
                "Statement": [{
                    "Effect": "Allow",
                    "Action": [
                        "cloudformation:DescribeStackEvents",
                        "cloudformation:DescribeStackResource",
                        "cloudformation:DescribeStackResources",
                        "cloudformation:DescribeStacks",
                        "cloudformation:ListStacks",
                        "cloudformation:ListStackResources"],
                    "Resource": "*"}]
            })

    def register_elb_to_dns(self,
                            elb,
                            tier_name,
                            tier_args):
        """
        Method handles the process of uniformly creating CNAME records for ELBs in a given tier
        @param elb [Troposphere.elasticloadbalancing.LoadBalancer]
        @param tier_name [str]
        @param tier_args [dict]
        """
        if 'environmentHostedZone' not in self.parameters:
            hostedzone = self.add_parameter(Parameter(
                "environmentHostedZone",
                Description="The DNS name of an existing Amazon Route 53 hosted zone",
                Default=tier_args.get('base_hosted_zone_name', 'devopsdemo.com'),
                Type="String"))
        else:
            hostedzone = self.parameters.get('environmentHostedZone')

        if tier_name.lower() + 'HostName' not in self.parameters:
            host_name = self.add_parameter(Parameter(
                tier_name.lower() + 'HostName',
                Description="Friendly host name to append to the environmentHostedZone base DNS record",
                Type="String",
                Default=tier_args.get('tier_host_name', tier_name.lower())))
        else:
            host_name = self.parameters.get(tier_name.lower() + 'HostName')

        self.add_resource(r53.RecordSetType(
            tier_name.lower() + 'DnsRecord',
            HostedZoneName=Join('', [Ref(hostedzone), '.']),
            Comment='CNAME record for ' + tier_name.capitalize() + ' tier',
            Name=Join('', [Ref(host_name), '.', Ref(hostedzone)]),
            Type='CNAME',
            TTL='300',
            ResourceRecords=[GetAtt(elb, 'DNSName')]))

    def get_logging_bucket_policy_document(self,
                                           utility_bucket,
                                           elb_log_prefix='elb_logs',
                                           cloudtrail_log_prefix='cloudtrail_logs'):
        """
        Method builds the S3 bucket policy statements which will allow the proper AWS account ids to write ELB Access Logs to the
        specified bucket and prefix.
        Per documentation located at: http://docs.aws.amazon.com/ElasticLoadBalancing/latest/DeveloperGuide/configure-s3-bucket.html
        @param utility_bucket [Troposphere.s3.Bucket] object reference of the utility bucket for this tier
        @param elb_log_prefix [string] prefix for paths used to prefix the path where ELB will place access logs
        """
        if elb_log_prefix is not None and elb_log_prefix != '':
            elb_log_prefix = elb_log_prefix + '/'
        else:
            elb_log_prefix = ''

        if cloudtrail_log_prefix is not None and cloudtrail_log_prefix != '':
            cloudtrail_log_prefix = cloudtrail_log_prefix + '/'
        else:
            cloudtrail_log_prefix = ''

        elb_accts = {'us-west-1': '027434742980',
                     'us-west-2': '797873946194',
                     'us-east-1': '127311923021',
                     'eu-west-1': '156460612806',
                     'ap-northeast-1': '582318560864',
                     'ap-southeast-1': '114774131450',
                     'ap-southeast-2': '783225319266',
                     'sa-east-1': '507241528517',
                     'us-gov-west-1': '048591011584'}

        for region in elb_accts:
            self.add_region_map_value(region, 'elbAccountId', elb_accts[region])

        # The principal account IDs in the following statements refer to the AWS CloudTrail account IDs
        # They explicitly need write permissions in order to upload logs to your bucket
        statements = [{
            "Action": ["s3:PutObject"],
            "Effect": "Allow",
            "Resource": Join('', ['arn:aws:s3:::', utility_bucket, '/', elb_log_prefix + 'AWSLogs/', Ref('AWS::AccountId'), '/*']),
            "Principal": {"AWS": [FindInMap('RegionMap', Ref('AWS::Region'), 'elbAccountId')]}},
            {
                "Action": ["s3:GetBucketAcl"],
                "Resource": Join('', ["arn:aws:s3:::", utility_bucket]),
                "Effect": "Allow",
                "Principal": {
                    "AWS": [
                        "arn:aws:iam::903692715234:root",
                        "arn:aws:iam::859597730677:root",
                        "arn:aws:iam::814480443879:root",
                        "arn:aws:iam::216624486486:root",
                        "arn:aws:iam::086441151436:root",
                        "arn:aws:iam::388731089494:root",
                        "arn:aws:iam::284668455005:root",
                        "arn:aws:iam::113285607260:root"]}},
            {
                "Action": ["s3:PutObject"],
                "Resource": Join('', [
                    "arn:aws:s3:::",
                    utility_bucket,
                    '/',
                    cloudtrail_log_prefix + "AWSLogs/",
                    Ref("AWS::AccountId"),
                    '/*'
                ]),
                "Effect": "Allow",
                "Principal": {
                    "AWS": [
                        "arn:aws:iam::903692715234:root",
                        "arn:aws:iam::859597730677:root",
                        "arn:aws:iam::814480443879:root",
                        "arn:aws:iam::216624486486:root",
                        "arn:aws:iam::086441151436:root",
                        "arn:aws:iam::388731089494:root",
                        "arn:aws:iam::284668455005:root",
                        "arn:aws:iam::113285607260:root"]},
                    "Condition": {"StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}}}]

        self.add_output(Output(
            'elbAccessLoggingBucketAndPath',
            Value=Join('', ['arn:aws:s3:::', utility_bucket, elb_log_prefix]),
            Description='S3 bucket and key name prefix to use when configuring elb access logs to aggregate to S3'))

        self.add_output(Output(
            'cloudTrailLoggingBucketAndPath',
            Value=Join('', ['arn:aws:s3:::', utility_bucket, cloudtrail_log_prefix]),
            Description='S3 bucket and key name prefix to use when configuring CloudTrail to aggregate logs to S3'))

        return {"Statement": statements}

    def create_vpcflowlogs_role(self):
        flowlogs_policy = aws.Policy(
            Version="2012-10-17",
            Statement=[
                aws.Statement(
                    Sid="",
                    Effect=aws.Allow,
                    Resource=['*'],
                    Action=[awacs_logs.CreateLogGroup,
                            awacs_logs.CreateLogStream,
                            awacs_logs.PutLogEvents,
                            awacs_logs.DescribeLogGroups,
                            awacs_logs.DescribeLogStreams],
                )
            ]
        )

        flowlogs_trust_policy = aws.Policy(
            Version="2012-10-17",
            Statement=[make_simple_assume_statement("vpc-flow-logs.amazonaws.com")]
        )

        vpcflowlogs_role = iam.Role(
            'VPCFlowLogsIAMRole',
            AssumeRolePolicyDocument=flowlogs_trust_policy,
            Path='/',
            Policies=[
                iam.Policy(PolicyName='vpcflowlogs_policy', PolicyDocument=flowlogs_policy)
            ])

        return vpcflowlogs_role

    def add_utility_bucket(self, name=None):
        """
        Method adds a bucket to be used for infrastructure utility purposes such as backups
        @param name [str] friendly name to prepend to the CloudFormation asset name
        """

        if name:
            self._utility_bucket = self.add_parameter(Parameter(
                'utilityBucket',
                Description='Name of the S3 bucket used for infrastructure utility',
                Default=name,
                AllowedPattern=res.get_str('ascii_only'),
                MinLength=1,
                MaxLength=255,
                ConstraintDescription=res.get_str('ascii_only_message'),
                Type='String'))
        else:
            self._utility_bucket = self.add_resource(s3.Bucket(
                name.lower() + 'UtilityBucket',
                AccessControl=s3.BucketOwnerFullControl,
                DeletionPolicy=Retain))

            bucket_policy_statements = self.get_logging_bucket_policy_document(
                self.utility_bucket,
                elb_log_prefix=res.get_str('elb_log_prefix', ''),
                cloudtrail_log_prefix=res.get_str('cloudtrail_log_prefix', ''))

            self.add_resource(s3.BucketPolicy(
                name.lower() + 'UtilityBucketLoggingPolicy',
                Bucket=self.utility_bucket,
                PolicyDocument=bucket_policy_statements))

        self.add_output(Output('utilityBucket', Value=self.utility_bucket))

        self.manual_parameter_bindings['utilityBucket'] = self.utility_bucket

    def add_log_group(self):
        log_group_name = 'DefaultLogGroup'
        self.add_resource(logs.LogGroup(
            log_group_name,
            RetentionInDays=7
        ))

    def add_vpcflowlogs_role(self):
        self.add_resource(self.create_vpcflowlogs_role())

    def add_child_template(self, child_template, merge=False, depends_on=[], output_autowire=True, propagate_outputs=True):
        """
        Appends the template to a list of child templates nested under this one
        These will be processed all together at the end of the create process in process_child_templates()
        """
        child_template_entry = (child_template, merge, depends_on, output_autowire, propagate_outputs)
        self._child_templates.append(child_template_entry)
        return child_template

    def add_child_template_reference(self, template_name, template_url, stack_params={}, depends_on=[]):
        """
        Create a child stack from a template that is not defined in this environment
        Useful for B/G deployments
        NOTE: If the original stack was deployed with any parameters, you must provide stack_params with
              this deployment. You can use utility.get_stack_params_from_parent_template() to
              retrieve the parameters that it was originally deployed with
        """
        return self.add_stack(
            template_name=template_name,
            template_url=template_url,
            stack_params=stack_params,
            depends_on=depends_on)

    def process_child_templates(self):
        """
        Iterate through and process the generated child template list
        """
        for (child_template, merge, depends_on, output_autowire, propagate_outputs) in self._child_templates:
            self.process_child_template(child_template, merge, depends_on, output_autowire, propagate_outputs)

    def process_child_template(self, child_template, merge, depends_on, output_autowire=True, propagate_outputs=True):
        """
        Add the common parameters from this template to the child template
        Execute the child template's build hook function
        Get the matching stack parameter values from this template and add the
        stack reference to this template with those stack parameters
        """

        # This merges all attributes from the two stacks together, so all this parameter binding is unnecessary
        if merge:
            self.merge(child_template)
            return

        # Add parameters from parent stack before executing build_hook
        child_template.add_common_parameters_from_parent(self)
        child_template.build_hook()
        if output_autowire:
            self.add_child_outputs_to_parameter_binding(child_template, propagate_up=propagate_outputs)

        # Match the stack parameters with parent stack parameter values and manual parameter bindings
        stack_params = self.match_stack_parameters(child_template)

        # Construct the resource path based on the prefix + name + timestamp
        child_template.resource_path = utility.get_template_s3_resource_path(
            prefix=Template.s3_path_prefix,
            template_name=child_template.name,
            include_timestamp=Template.include_timestamp)

        # Construct the template url using the bucket name and resource path
        template_s3_url = self.get_template_s3_url(child_template)

        # Create the stack resource in this template and return the reference
        return self.add_stack(
            template_name=child_template.name,
            template_url=template_s3_url,
            stack_params=stack_params,
            depends_on=depends_on)

    def get_template_s3_url(self, child_template):
        """
        Overridable method for getting the s3 url for child templates.

        By default it uses the `TemplateBucket` Parameter and
            `child_template.resource_path` to build the URL.
        Use `utility.get_template_s3_url(Template.template_bucket_default, child_template.resource_path)`
            if you want a non-parametrized version of this URL.
        """
        return Join('', ['https://', Ref(self.template_bucket_param), '.s3.amazonaws.com/', child_template.resource_path])

    def add_stack(self, template_name, template_url, stack_params={}, depends_on=[]):
        """
        Creates a cloudformation stack resource in this template with the attributes provided
        """
        stack_obj = cf.Stack(
            template_name,
            TemplateURL=template_url,
            Parameters=stack_params,
            TimeoutInMinutes=Template.stack_timeout,
            DependsOn=depends_on)

        return self.add_resource(stack_obj)

    def add_child_outputs_to_parameter_binding(self, child_template, propagate_up=False):
        """
        This auto-wires the outputs of the child stack to the manual_param of the parent stack
        """
        for output in child_template.outputs:
            self.manual_parameter_bindings[output] = GetAtt(child_template.name, "Outputs." + output)
            if propagate_up:
                self.add_output(Output(output, Value=GetAtt(child_template.name, "Outputs." + output)))
            # TODO: should a custom resource be addeded for each output?

    def match_stack_parameters(self, child_template):
        """
        For all matching parameters between this template and the child template, attempt to
        get the value from this template's parameters, resources, and manual parameter bindings.
        Return the dictionary of stack parameters to deploy the child template with
        """
        stack_params = {}

        for parameter in child_template.parameters.keys():

            # Manual parameter bindings single-namespace
            if parameter in self.manual_parameter_bindings:
                manual_match = self.manual_parameter_bindings[parameter]
                stack_params[parameter] = manual_match

            # Match any child stack parameters that have the same name as this stacks **parameters**
            elif parameter in self.parameters.keys():
                param_match = self.parameters.get(parameter)
                stack_params[parameter] = Ref(param_match)

            # Match any child stack parameters that have the same name as this stacks **resources**
            elif parameter in self.resources.keys():
                resource_match = self.resources.get(parameter)
                stack_params[parameter] = Ref(resource_match)

            # # Match any child stack parameters that have the same name as a top-level **stack_output**
            # TODO: Enable Output autowiring
            # elif parameter in self.stack_outputs:
            #     stack_params[parameter] = GetAtt(self.stack_outputs[parameter], 'Outputs.' + parameter)

            # Finally if nothing else matches copy the child templates parameter to this template's parameter list
            # so the value will pass through this stack down to the child.
            else:
                new_param = self.add_parameter(child_template.parameters[parameter])
                stack_params[parameter] = Ref(new_param)

        return stack_params

    def get_subnet_type(self, subnet_layer):
        """
        Return the subnet type (public/private) that subnet_layer belongs to
        """
        for subnet_type in self._subnets:
            for a_subnet_layer in self._subnets[subnet_type]:
                if a_subnet_layer == subnet_layer:
                    return subnet_type
        return None



from environmentbase.template import Template, DEFAULT_TO_MIN_SIZE
from troposphere import Ref, Parameter, Output, GetAtt, ec2, route53, autoscaling
from troposphere.policies import CreationPolicy, ResourceSignal
from troposphere.policies import UpdatePolicy, AutoScalingRollingUpdate

SCHEME_INTERNET_FACING = 'internet-facing'
SCHEME_INTERNAL = 'internal'
PUBLIC_ACCESS_CIDR = '0.0.0.0/0'


class HaCluster(Template):
    """
    Generic highly available cluster template
    Contains an ELB, Autoscaling Group, Security Groups, and optional internal DNS record
    """

    def __init__(self,
                 name='HaCluster',
                 ami_name='amazonLinuxAmiId',
                 user_data='',
                 env_vars={},
                 min_size=1,
                 max_size=1,
                 desired_capacity=DEFAULT_TO_MIN_SIZE,
                 instance_type='t2.micro',
                 subnet_layer=None,
                 elb_scheme=SCHEME_INTERNET_FACING,
                 elb_listeners=[{'elb_protocol': 'HTTP', 'elb_port': 80}],
                 elb_health_check_port=None,
                 elb_health_check_protocol='TCP',
                 elb_health_check_path='',
                 elb_idle_timeout=None,
                 update_policy_PauseTime='PT1M',
                 update_policy_MinInstancesInService=0,
                 update_policy_MaxBatchSize=1,
                 cname='',
                 custom_tags={},
                 elb_custom_tags={},
                 scaling_policies=None,
                 creation_policy_timeout=None,
                 allow_default_ingress=True):

        # This will be the name used in resource names and descriptions
        self.name = name

        # This is the name used to identify the AMI from the ami_cache.json file
        self.ami_name = ami_name

        # This is the contents of the userdata script as a string
        self.user_data = user_data

        # This is a dictionary of environment variables to inject into the instances
        self.env_vars = env_vars

        # These define the lower and upper boundaries of the autoscaling group
        self.min_size = min_size
        self.max_size = max_size
        self.desired_capacity = desired_capacity

        # The type of instance for the autoscaling group
        self.instance_type = instance_type

        # This is the subnet layer that the ASG is in (public, private, ...)
        self.subnet_layer = subnet_layer

        # This is the type of ELB: internet-facing gets a publicly accessible DNS, while internal is only accessible to the VPC
        self.elb_scheme = elb_scheme

        # This should be a list of dictionaries defining each listener for the ELB
        # Each dictionary can contain elb_port [required], elb_protocol, instance_port, instance_protocol, ssl_cert_name
        self.elb_listeners = elb_listeners

        # This is the health check port for the cluster
        self.elb_health_check_port = elb_health_check_port

        # The ELB health check protocol for the cluster (HTTP, HTTPS, TCP, SSL)
        self.elb_health_check_protocol = elb_health_check_protocol

        # The ELB health check path for the cluster (Only for HTTP and HTTPS)
        self.elb_health_check_path = elb_health_check_path

        # Add a creation policy with a custom timeout if one was specified
        if creation_policy_timeout:
            self.creation_policy = CreationPolicy(ResourceSignal=ResourceSignal(Timeout='PT' + str(creation_policy_timeout) + 'M'))
        else:
            self.creation_policy = None

        # Add update policy
        self.update_policy = UpdatePolicy(
            AutoScalingRollingUpdate=AutoScalingRollingUpdate(
                PauseTime=update_policy_PauseTime,
                MinInstancesInService=update_policy_MinInstancesInService,
                MaxBatchSize=update_policy_MaxBatchSize,
                # WaitOnResourceSignals=True
            )
        )

        # The Idle Timeout for the ELB (how long your connection can stay idle before being terminated)
        self.elb_idle_timeout = elb_idle_timeout

        # This is an optional fully qualified DNS name to create a CNAME in a private hosted zone
        self.cname = cname

        # Translate the custom_tags dict to a list of autoscaling Tags
        self.custom_tags = []
        for key, value in custom_tags.iteritems():
            self.custom_tags.append(autoscaling.Tag(key, value, True))

        # Save ELB tags for add_cluster_elb
        self.elb_custom_tags = elb_custom_tags

        # A list of dictionaries describing scaling policies to be passed to add_asg
        self.scaling_policies = scaling_policies

        # Indicates whether ingress rules should be added to the ELB for type-appropriate CIDR ranges
        # Internet facing ELBs would allow ingress from PUBLIC_ACCESS_CIDR and private ELBs will allow ingress from the VPC CIDR
        self.allow_default_ingress = allow_default_ingress

        super(HaCluster, self).__init__(template_name=self.name)

    def build_hook(self):
        """
        Hook to add tier-specific assets within the build stage of initializing this class.
        """

        # Set the subnet_layer if it wasn't passed in
        self.set_subnet_layer()

        # Create security groups for the ASG and ELB and connect them together
        self.add_security_groups()

        # Add the IAM role for the autoscaling group
        self.add_cluster_instance_profile()

        # Add the ELB for the autoscaling group
        self.add_cluster_elb()

        # Add the CNAME for the ELB
        self.add_cname()

        # Add the userdata for the autoscaling group
        self.add_user_data()

        # Add the autoscaling group for the cluster
        self.add_cluster_asg()

        # Add the outputs for the stack
        self.add_outputs()

    def set_subnet_layer(self):
        """
        If the subnet layer is not passed in, use a private subnet if there are any, otherwise use a public subnet.
        This needs to happen in the build hook, since subnets is not yet initialized in the constructor. You
        probably won't need to override this. This logic is also duplicated in template.add_asg(), but we need to
        set it out here so we can pass the same subnet to template.add_elb()
        """
        if not self.subnet_layer:
            if len(self._subnets.get('private')) > 0:
                self.subnet_layer = self._subnets['private'].keys()[0]
            else:
                self.subnet_layer = self._subnets['public'].keys()[0]

    def add_security_groups(self):
        """
        Wrapper method to encapsulate process of creating security groups for this tier
        Creates security groups for both ASG and ELB and opens the ports between them
        Sets self.security_groups as a dictionary with two security_groups: ha_cluster and elb
        """

        elb_sg_ingress_rules = []

        if self.allow_default_ingress:
            # Determine ingress rules for ELB security -- open to internet for internet-facing ELB, open to VPC for internal ELB
            access_cidr = PUBLIC_ACCESS_CIDR if self.elb_scheme == SCHEME_INTERNET_FACING else self.vpc_cidr

            # Add the ingress rules to the ELB security group
            for elb_port in [listener.get('elb_port') for listener in self.elb_listeners]:
                elb_sg_ingress_rules.append(ec2.SecurityGroupRule(FromPort=elb_port, ToPort=elb_port, IpProtocol='tcp', CidrIp=access_cidr))

        # Create the ELB security group and attach the ingress rules
        elb_sg_name = '%sElbSecurityGroup' % self.name
        elb_sg = self.add_resource(
            ec2.SecurityGroup(
                elb_sg_name,
                GroupDescription='Security group for %s ELB' % self.name,
                VpcId=self.vpc_id,
                SecurityGroupIngress=elb_sg_ingress_rules)
        )

        # Create the ASG security group
        ha_cluster_sg_name = '%sSecurityGroup' % self.name
        ha_cluster_sg = self.add_resource(
            ec2.SecurityGroup(
                ha_cluster_sg_name,
                GroupDescription='Security group for %s' % self.name,
                VpcId=self.vpc_id)
        )

        # Create the reciprocal rules between the ELB and the ASG for all instance ports
        # NOTE: The condition in the list comprehension exists because elb_port is used as a default when instance_port is not specified
        cluster_sg_ingress_ports = {listener.get('instance_port') if listener.get('instance_port') else listener.get('elb_port')
                                    for listener in self.elb_listeners}

        # Also add the health check port to the security group rules
        if self.elb_health_check_port:
            cluster_sg_ingress_ports.add(self.elb_health_check_port)

        for cluster_sg_ingress_port in cluster_sg_ingress_ports:
            self.create_reciprocal_sg(
                elb_sg, elb_sg_name,
                ha_cluster_sg, ha_cluster_sg_name,
                from_port=cluster_sg_ingress_port)

        self.security_groups = {'ha_cluster': ha_cluster_sg, 'elb': elb_sg}

        return self.security_groups

    def add_cluster_instance_profile(self):
        """
        Wrapper method to encapsulate process of adding the IAM role for the autoscaling group
        Sets self.instance_profile with the IAM Role resource, used in creation of the Launch Configuration
        """
        self.instance_profile = None

    def add_cluster_elb(self):
        """
        Wrapper method to encapsulate process of creating the ELB for the autoscaling group
        Sets self.cluster_elb with the ELB resource
        """
        # If the ELB is internal, use the same subnet layer as the ASG
        elb_subnet_layer = self.subnet_layer if self.elb_scheme == SCHEME_INTERNAL else None

        # This creates the ELB, opens the specified ports, and attaches the security group and logging bucket
        self.cluster_elb = self.add_elb(
            resource_name=self.name,
            security_groups=[self.security_groups['elb']],
            listeners=self.elb_listeners,
            elb_custom_tags=self.elb_custom_tags,
            utility_bucket=self.utility_bucket,
            subnet_layer=elb_subnet_layer,
            scheme=self.elb_scheme,
            idle_timeout=self.elb_idle_timeout,
            health_check_port=self.elb_health_check_port,
            health_check_protocol=self.elb_health_check_protocol,
            health_check_path=self.elb_health_check_path
        )

    def add_cname(self):
        """
        Wrapper method to encapsulate process of creating a CNAME DNS record for the ELB
        Requires InternalHostedZone parameter
        Sets self.cname_record with the record resource
        """

        if not self.cname:
            return

        hosted_zone = self.add_parameter(Parameter(
            'InternalHostedZone',
            Description='Internal Hosted Zone Name',
            Type='String'))

        self.cname_record = self.add_resource(route53.RecordSetType(
            self.name.lower() + 'DnsRecord',
            HostedZoneId=Ref(hosted_zone),
            Comment='CNAME record for %s' % self.name,
            Name=self.cname,
            Type='CNAME',
            TTL='300',
            ResourceRecords=[GetAtt(self.cluster_elb, 'DNSName')]))

    def add_user_data(self):
        """
        Wrapper method to encapsulate process of constructing userdata for the autoscaling group
        Sets self.user_data_payload constructed from the passed in user_data and env_vars
        """
        self.user_data_payload = self.construct_user_data(self.env_vars, self.user_data)

    def add_cluster_asg(self):
        """
        Wrapper method to encapsulate process of creating the autoscaling group
        Sets self.cluster_asg with the autoscaling group resource
        """
        self.cluster_asg = self.add_asg(
            layer_name=self.name,
            security_groups=[self.security_groups['ha_cluster'], self.common_security_group],
            load_balancer=self.cluster_elb,
            ami_name=self.ami_name,
            user_data=self.user_data_payload,
            instance_type=self.instance_type,
            min_size=self.min_size,
            max_size=self.max_size,
            desired_capacity=self.desired_capacity,
            subnet_layer=self.subnet_layer,
            instance_profile=self.instance_profile,
            custom_tags=self.custom_tags,
            creation_policy=self.creation_policy,
            update_policy=self.update_policy,
            scaling_policies=self.scaling_policies
        )

    def add_outputs(self):
        """
        Wrapper method to encapsulate creation of stack outputs for this template
        """
        self.add_output(Output('%sELBDNSName' % self.name, Value=GetAtt(self.cluster_elb, 'DNSName')))
        self.add_output(Output('%sSecurityGroupId' % self.name, Value=Ref(self.security_groups['ha_cluster'])))
        self.add_output(Output('%sElbSecurityGroupId' % self.name, Value=Ref(self.security_groups['elb'])))



from environmentbase.template import Template
from environmentbase import resources
from troposphere import Ref, Join, Base64, FindInMap, Output
from troposphere.ec2 import SecurityGroup, SecurityGroupIngress, SecurityGroupEgress
from troposphere.autoscaling import AutoScalingGroup, LaunchConfiguration, Tag
from troposphere.iam import Policy, Role, InstanceProfile
from troposphere.policies import CreationPolicy, ResourceSignal


class HaNat(Template):
    '''
    Adds a highly available NAT that also serves as an NTP server
    Creates a 1-1 autoscaling group in the provided public subnet and creates
    a route directing egress traffic from the private subnet through this NAT
    '''

    def __init__(self, subnet_index, instance_type='t2.micro', enable_ntp=False, name='HaNat', extra_user_data=None):
        '''
        Method initializes HA NAT in a given environment deployment
        @param subnet_index [int] ID of the subnet that the NAT instance will be deployed to
        @param instance_type [string] - Type of NAT instance in the autoscaling group
        '''
        self.subnet_index = subnet_index
        self.instance_type = instance_type
        self.enable_ntp = enable_ntp
        self.extra_user_data = extra_user_data

        # These will be initialized and consumed by various functions called in the build hook
        self.sg = None
        self.instance_profile = None

        super(HaNat, self).__init__(template_name=name)

    def build_hook(self):
        '''
        Hook to add tier-specific assets within the build stage of initializing this class.
        '''
        self.add_nat_sg()
        self.add_nat_instance_profile()
        self.add_nat_asg()

    def add_nat_sg(self):
        '''
        Create the NAT security group and add the ingress/egress rules
        '''
        self.sg = self.add_resource(SecurityGroup(
            "Nat%sSG" % str(self.subnet_index),
            VpcId=self.vpc_id,
            GroupDescription="Security group for NAT host."
        ))
        self.add_output(Output(self.sg.name, Value=Ref(self.sg)))
        self.add_nat_sg_rules()

    def add_nat_sg_rules(self):
        '''
        Add the security group rules necessary for the NAT to operate
        For now, this is opening all ingress from the VPC and all egress to the internet
        '''
        self.add_resource(SecurityGroupIngress(
            "Nat%sIngress" % str(self.subnet_index),
            ToPort="-1",
            FromPort="-1",
            IpProtocol="-1",
            GroupId=Ref(self.sg),
            CidrIp=self.vpc_cidr
        ))
        self.add_resource(SecurityGroupEgress(
            "Nat%sEgress" % str(self.subnet_index),
            ToPort="-1",
            FromPort="-1",
            IpProtocol="-1",
            GroupId=Ref(self.sg),
            CidrIp='0.0.0.0/0'
        ))

    def get_extra_policy_statements(self):
        '''
        Return policy statements for additional permissions you want your NAT
        to have.
        '''
        return []

    def add_nat_instance_profile(self):
        '''
        Create the NAT role and instance profile
        '''
        policy_actions = [
            "ec2:DescribeInstances",
            "ec2:ModifyInstanceAttribute",
            "ec2:DescribeSubnets",
            "ec2:DescribeRouteTables",
            "ec2:CreateRoute",
            "ec2:ReplaceRoute",
            "ec2:StartInstances",
            "ec2:StopInstances"
        ]
        if self.enable_ntp:
            policy_actions.extend([
                "ec2:*DhcpOptions*",
                "ec2:DescribeVpcs"
            ])

        nat_role = self.add_resource(
            Role(
                "Nat%sRole" % str(self.subnet_index),
                AssumeRolePolicyDocument={
                    "Statement": [{
                        "Effect": "Allow",
                        "Principal": {
                            "Service": ["ec2.amazonaws.com"]
                        },
                        "Action": ["sts:AssumeRole"]
                    }]
                },
                Path="/",
                Policies=[
                    Policy(
                        PolicyName="NAT%sPolicy" % str(self.subnet_index),
                        PolicyDocument={
                            "Statement": [{
                                "Effect": "Allow",
                                "Action": policy_actions,
                                "Resource": "*"
                            }] + self.get_extra_policy_statements()
                        }
                    )]
            ))

        self.instance_profile = self.add_resource(InstanceProfile(
            "Nat%sInstanceProfile" % str(self.subnet_index),
            Path="/",
            Roles=[Ref(nat_role)]
        ))

    def add_nat_asg(self):

        user_data = [resources.get_resource('nat_takeover.sh')]

        if self.enable_ntp:
            user_data.append(resources.get_resource('ntp_takeover.sh'))
        if self.extra_user_data:
            user_data.append(open(self.extra_user_data).read())

        nat_asg_name = "Nat%sASG" % str(self.subnet_index)

        user_data.extend([
            "\n",
            "cfn-signal -s true",
            " --resource ", nat_asg_name,
            " --stack ", {"Ref": "AWS::StackName"},
            " --region ", {"Ref": "AWS::Region"}
        ])

        nat_launch_config = self.add_resource(LaunchConfiguration(
            "Nat%sLaunchConfig" % str(self.subnet_index),
            UserData=Base64(Join('', user_data)),
            ImageId=FindInMap('RegionMap', Ref('AWS::Region'), 'natAmiId'),
            KeyName=Ref('ec2Key'),
            SecurityGroups=[Ref(self.sg)],
            EbsOptimized=False,
            IamInstanceProfile=Ref(self.instance_profile),
            InstanceType=self.instance_type,
            AssociatePublicIpAddress=True
        ))

        # Create the NAT in a public subnet
        subnet_layer = self._subnets['public'].keys()[0]

        nat_asg = self.add_resource(AutoScalingGroup(
            nat_asg_name,
            DesiredCapacity=1,
            Tags=[
                Tag("Name", Join("-", ["NAT", self.subnet_index, ]), True),
                Tag("isNat", "true", True)
            ],
            MinSize=1,
            MaxSize=1,
            Cooldown="30",
            LaunchConfigurationName=Ref(nat_launch_config),
            HealthCheckGracePeriod=30,
            HealthCheckType="EC2",
            VPCZoneIdentifier=[self._subnets['public'][subnet_layer][self.subnet_index]],
            CreationPolicy=CreationPolicy(
                ResourceSignal=ResourceSignal(
                    Count=1,
                    Timeout='PT15M'
                )
            )
        ))

        return nat_asg



from troposphere import GetAtt
from .environmentbase import EnvironmentBase
from .patterns.base_network import BaseNetwork


class NetworkBase(EnvironmentBase):
    """
    EnvironmentBase controller containing a root template with all of the base networking infrastructure
    for a common deployment within AWS. This is intended to be the 'base' stack for deploying child stacks
    """

    def add_config_hook(self):
        super(NetworkBase, self).add_config_hook()
        self._add_config_handler(BaseNetwork)

    def create_hook(self):
        super(NetworkBase, self).create_hook()

        network_config = self.config.get('network', {})
        nat_config = self.config.get('nat')

        base_network_template = BaseNetwork('BaseNetwork', network_config, nat_config)
        self.add_child_template(base_network_template)

        self.template._subnets = base_network_template._subnets.copy()
        self.template._vpc_id = GetAtt(base_network_template.name, 'Outputs.vpcId')



#!/usr/bin/env python
"""awsbootstrap.py

This script handles overall AWS Account setup and structure for globally-enabled, non-deployment environment assets (such as CloudTrail).

Usage:
    awsbootstrap.py [--existing_bucket <EXISTING_BUCKET>] [--bucket_region <BUCKET_REGION>] [--region <REGION>]
                    [--generate_topics] [--topic_name <TOPIC_NAME>] [--trail_name <TRAIL_NAME>]
                    [--third_party_auth_ids] [--debug]

Options:
  -h --help                            Show this screen.
  -v --version                         Show version.
  --debug                              Prints parent template to console out [default: 0].
  --existing_bucket <EXISTING_BUCKET>  Indicates that an existing bucket should be used.
  --bucket_region <BUCKET_REGION>      Region in which to create the S3 bucket for cloudtrail aggregation [default: us-west-2].
  --generate_topics                    Command-line switch indicating whether topics will be generated or not [default: 0].
  --topic_name <TOPIC_NAME>            Name of the topic to create when generating SNS topics for CloudTrail [deault: cloudtrailtopic].
  --trail_name <TRAIL_NAME>            Name of the trail to create within CloudTrail [default: Default].
  --region <REGION>                    Comma separated list of regions to apply this setup to [default: all].
  --stack_name <STACK_NAME>            User-definable value for the CloudFormation stack being deployed [default: accountBootstrapStack].
  --third_party_auth_ids               Command-line switch indicating whether an API credential will be generated or not [default: 0].
"""

from docopt import docopt
import logging
import boto
import boto.iam
import boto.sns
import boto.cloudtrail
import boto.cloudformation
import json
from troposphere import Template, Ref, Join, GetAtt, Output
import troposphere.iam as iam
import troposphere.s3 as s3
from environmentutil import EnvironmentUtil

arguments = docopt(__doc__, version='devtools 0.1')

if arguments['--debug']:
    level = logging.DEBUG
else:
    level = logging.INFO

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=level)

regions = []
if arguments.get('--region', 'all') == 'all':
    for region in boto.vpc.regions():
        regions.append(region.name)
else:
    for region in arguments.get('--region', 'all').split(','):
        regions.append(region)

env_util = EnvironmentUtil({})

t = Template()

if arguments.get('--third_parth_auth_ids', False):
    federated_auth_user = t.add_resource(iam.User(
        'federatedAuthUser',
        Path='/federatedauth/',
        Policies=[
            iam.Policy(
                PolicyName='federationUserIAMGetList',
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [{
                        "Sid": "Stmt1429206486000",
                        "Effect": "Allow",
                        "Action": [
                            "iam:Get*",
                            "iam:List*"],
                        "Resource": [
                            "*"]}]})]))

    t.add_output(Output(
        'federatedAuthUser',
        Value=Ref(federated_auth_user),
        Description='Name of the AWS user that federated auth system will authenticate through.'))

    t.add_output(Output(
        'federatedAuthUserArn',
        Value=GetAtt(federated_auth_user, 'Arn')))

    assume_role_policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "AWS": Join('', ['', GetAtt(federated_auth_user, 'Arn'), ''])
                },
                "Action": "sts:AssumeRole"
            }
        ]
    }

    federated_auth_api_keys = t.add_resource(iam.AccessKey(
        'federatedAuthAccessKey',
        UserName=Ref(federated_auth_user)))

if arguments.get('--existing_bucket', None):
    logging_bucket = arguments.get('--existing_bucket', None)
else:
    logging_bucket_resource = t.add_resource(s3.Bucket(
        'cloudTrailBucket',
        DependsOn=federated_auth_user.title))
    logging_bucket = Ref(logging_bucket_resource)

t.add_output(Output(
    'bucketName',
    Value=logging_bucket,
    Description='Name of the S3 bucket created for logging'))

role_base = {
    'DependsOn': logging_bucket_resource.title,
    'AssumeRolePolicyDocument': assume_role_policy_document,
    'Path': '/userroles/',
    'Policies': []
}

admin_args = role_base.copy()
admin_args['Policies'].append(
    iam.Policy(
        PolicyName='adminAccess',
        PolicyDocument={
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "*"}]}))

admin_role = t.add_resource(iam.Role('adminRole', **admin_args))

t.add_output(Output(
    'adminRole',
    Value=Ref(admin_role),
    Description='Name of the AWS role for administrative access to the Console and API'))

developer_args = role_base.copy()
developer_args['Policies'].append(
    iam.Policy(
        PolicyName='powerUserAccess',
        PolicyDocument={
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "NotAction": "iam:*",
                    "Resource": "*"
                }
            ]
        }
    )
)

developer_role = t.add_resource(iam.Role('developerRole', **developer_args))

t.add_output(Output(
    'developerRole',
    Value=Ref(developer_role),
    Description='Name of the AWS role for developer or power user access to the Console and API'))

read_only_iam_policy = iam.Policy(
    PolicyName='readOnlyAccess',
    PolicyDocument={
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": [
                    "appstream:Get*",
                    "autoscaling:Describe*",
                    "cloudformation:DescribeStacks",
                    "cloudformation:DescribeStackEvents",
                    "cloudformation:DescribeStackResource",
                    "cloudformation:DescribeStackResources",
                    "cloudformation:GetTemplate",
                    "cloudformation:List*",
                    "cloudfront:Get*",
                    "cloudfront:List*",
                    "cloudtrail:DescribeTrails",
                    "cloudtrail:GetTrailStatus",
                    "cloudwatch:Describe*",
                    "cloudwatch:Get*",
                    "cloudwatch:List*",
                    "directconnect:Describe*",
                    "dynamodb:GetItem",
                    "dynamodb:BatchGetItem",
                    "dynamodb:Query",
                    "dynamodb:Scan",
                    "dynamodb:DescribeTable",
                    "dynamodb:ListTables",
                    "ec2:Describe*",
                    "elasticache:Describe*",
                    "elasticbeanstalk:Check*",
                    "elasticbeanstalk:Describe*",
                    "elasticbeanstalk:List*",
                    "elasticbeanstalk:RequestEnvironmentInfo",
                    "elasticbeanstalk:RetrieveEnvironmentInfo",
                    "elasticloadbalancing:Describe*",
                    "elasticmapreduce:Describe*",
                    "elasticmapreduce:List*",
                    "elastictranscoder:Read*",
                    "elastictranscoder:List*",
                    "iam:List*",
                    "iam:Get*",
                    "kinesis:Describe*",
                    "kinesis:Get*",
                    "kinesis:List*",
                    "opsworks:Describe*",
                    "opsworks:Get*",
                    "route53:Get*",
                    "route53:List*",
                    "redshift:Describe*",
                    "redshift:ViewQueriesInConsole",
                    "rds:Describe*",
                    "rds:ListTagsForResource",
                    "s3:Get*",
                    "s3:List*",
                    "sdb:GetAttributes",
                    "sdb:List*",
                    "sdb:Select*",
                    "ses:Get*",
                    "ses:List*",
                    "sns:Get*",
                    "sns:List*",
                    "sqs:GetQueueAttributes",
                    "sqs:ListQueues",
                    "sqs:ReceiveMessage",
                    "storagegateway:List*",
                    "storagegateway:Describe*",
                    "tag:get*",
                    "trustedadvisor:Describe*"
                ],
                "Effect": "Allow",
                "Resource": "*"
            }
        ]
    })


read_only_role = t.add_resource(iam.Role(
    'readOnlyRole',
    DependsOn=logging_bucket_resource.title,
    AssumeRolePolicyDocument=assume_role_policy_document,
    Path='/userroles/',
    Policies=[read_only_iam_policy]))

t.add_output(Output(
    'readOnlyRole',
    Value=Ref(read_only_role),
    Description='Name of the AWS role for read only access to the Console and API'))

if arguments.get('--third_parth_auth_ids', False):

    t.add_output(Output(
        'federatedAuthUserAccessKeyId',
        Description='AWS Access Key ID for federated auth user',
        Value=Ref(federated_auth_api_keys)))

    t.add_output(Output(
        'federatedAuthUserSecretAccessKey',
        Description='AWS Secret Access Key for federated auth user',
        Value=GetAtt(federated_auth_api_keys, 'SecretAccessKey')))

    logging.debug('**********')
    logging.debug('Intermediate template (with creds in output): ' + t.to_json())
    logging.debug('**********')

    cfconn = boto.cloudformation.connect_to_region(arguments.get('--bucket_region', 'us-west-2'))
    logging.info('Connected to CloudFormation in region: ' + arguments.get('--bucket_region>', 'us-west-2'))
    logging.info('Starting deploy of intermediate template to gather federated authentication IAM credentials')

    cfconn.create_stack(
        arguments.get('--stack_name', 'accountBootstrapStack'),
        template_body=t.to_json(),
        capabilities=['CAPABILITY_IAM'])

    env_util.wait_for_stack(cfconn,
                            arguments.get('--stack_name', 'accountBootstrapStack'))

    stack = cfconn.describe_stacks(stack_name_or_id=arguments.get('--stack_name', 'accountBootstrapStack'))[0]
    logging.info('Stack %s has completely deployed with status of %s' % (arguments.get('--stack_name', 'accountBootstrapStack'),
                                                                         stack.stack_status))

    output_variables = {}

    if stack.stack_status == 'CREATE_COMPLETE':
        logging.info('Intermediate Stack has deployed with status ' + stack.stack_status)
        for output in stack.outputs:
            output_variables[output.key] = output.value

    del t.outputs['federatedAuthUserAccessKeyId']
    del t.outputs['federatedAuthUserSecretAccessKey']

    t.add_resource(
        s3.BucketPolicy(
            'cloudTrailBucketPolicy',
            Bucket=logging_bucket,
            PolicyDocument={
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "AWSCloudTrailAclCheck20131101",
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::903692715234:root",
                                "arn:aws:iam::859597730677:root",
                                "arn:aws:iam::814480443879:root",
                                "arn:aws:iam::216624486486:root",
                                "arn:aws:iam::086441151436:root",
                                "arn:aws:iam::388731089494:root",
                                "arn:aws:iam::284668455005:root",
                                "arn:aws:iam::113285607260:root",
                                "arn:aws:iam::035351147821:root"
                            ]
                        },
                        "Action": "s3:GetBucketAcl",
                        "Resource": Join('', ["arn:aws:s3:::", logging_bucket])
                    },
                    {
                        "Sid": "AWSCloudTrailWrite20131101",
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::903692715234:root",
                                "arn:aws:iam::859597730677:root",
                                "arn:aws:iam::814480443879:root",
                                "arn:aws:iam::216624486486:root",
                                "arn:aws:iam::086441151436:root",
                                "arn:aws:iam::388731089494:root",
                                "arn:aws:iam::284668455005:root",
                                "arn:aws:iam::113285607260:root",
                                "arn:aws:iam::035351147821:root"
                            ]
                        },
                        "Action": "s3:PutObject",
                        "Resource": Join('', ["arn:aws:s3:::", logging_bucket, "/AWSLogs/", Ref('AWS::AccountId'), "/*"]),
                        "Condition": {
                            "StringEquals": {
                                "s3:x-amz-acl": "bucket-owner-full-control"
                            }
                        }
                    }
                ]
            }))

    logging.debug('**********')
    logging.debug('Final template (without creds in output): ' + t.to_json())
    logging.debug('**********')

    logging.debug('Updating CloudFormation stack with final template')
    cfconn.update_stack(
        stack_name=arguments.get('--stack_name', 'accountBootstrapStack'),
        template_body=t.to_json(),
        capabilities=['CAPABILITY_IAM'])

    env_util.wait_for_stack(cfconn, arguments.get('--stack_name', 'accountBootstrapStack'))

    stack = cfconn.describe_stacks(stack_name_or_id=arguments.get('--stack_name', 'accountBootstrapStack'))[0]
    if stack.stack_status == 'UPDATE_COMPLETE':
        logging.info('Stack %s has completely deployed with status of %s' % (arguments.get('--stack_name', 'accountBootstrapStack'),
                                                                             stack.stack_status))
    else:
        logging.error('Final stack failed to deploy. Please check errors in AWS console, repair CloudFormation template and redeploy. Note'
                      ' that credentials are currently visible in the CloudFormation Console.')
        exit(1)
else:
    logging.error('Stack failed to deploy. Please check errors in AWS console, repair CloudFormation template and redeploy.')
    exit(1)

iamconn = boto.connect_iam()
output_variables['loginUrl'] = iamconn.get_signin_url()

logging.debug('Regions:' + json.dumps(regions))

global_logging = True

for aws_region in regions:
    if aws_region in ['cn-north-1', 'us-gov-west-1']:
        logging.debug('Ignoring restricted region: ' + aws_region)
    else:
        logging.info('********************************************************************************')
        logging.info(' Starting to create CloudTrail resources in region: ' + aws_region)
        if arguments['--generate_topics']:
            logging.info('Generating topic for CloudTrail in region: ' + aws_region)
            sns = boto.sns.connect_to_region(aws_region)
            logging.debug('Connected to SNS API in region: ' + aws_region)
            arguments['topic_name'] = sns.create_topic(arguments['--topic_name'])['CreateTopicResponse']['CreateTopicResult']['TopicArn']

        ct = boto.cloudtrail.connect_to_region(aws_region)
        logging.debug('Connected to CloudTrail API in region: ' + aws_region)
        if len(ct.describe_trails(trail_name_list=[arguments.get('--trail_name', 'Default')]).get('trailList', [])) == 0:
            logging.info('Creating new trail in region ' + aws_region)
            ct.create_trail(
                name=arguments.get('--trail_name', 'Default'),
                s3_bucket_name=output_variables.get('bucketName', ''),
                s3_key_prefix=arguments.get('--s3_key_prefix'),
                sns_topic_name=arguments.get('topic_name'),
                include_global_service_events=global_logging)
            ct.start_logging(name=arguments.get('--trail_name', 'Default'))
            if global_logging:
                global_logging = False
        elif not ct.get_trail_status('Default').get('IsLogging'):
            logging.info('Updating Trail in region ' + aws_region)
            ct.update_trail(
                name=arguments.get('--trail_name', 'Default'),
                s3_bucket_name=output_variables.get('bucketName', ''),
                s3_key_prefix=arguments.get('--s3_key_prefix'),
                sns_topic_name=arguments.get('topic_name'),
                include_global_service_events=global_logging)
            if global_logging:
                global_logging = False
            ct.start_logging(name=arguments.get('--trail_name', 'Default'))
        else:
            logging.info('Trail ' + arguments.get('--trail_name', 'Default') + ' already exists in region ' + aws_region)

logging.info('Bucket created for CloudTrail logs in %s with name of %s' % (arguments.get('--bucket_region', 'us-west-2'),
                                                                           output_variables.get('bucketName', '')))

print('\n\n\n\n')
logging.info('### Process Complete ###')
if arguments.get('--third_parth_auth_ids', False):
    logging.info('    Provide the following JSON document to the team managing federated auth for provisioning:')
    print('\n')
    print(json.dumps(output_variables))
print('\n\n\n\n')



import random
import string
import boto3
import json
import time
import troposphere as t
import tempfile
import os
from . import resources as res


def random_string(size=5):
    return ''.join(random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits) for _ in range(size))


def _get_boto_session(boto_config):
    if not boto_config.get('session'):
        boto_config['session'] = boto3.session.Session(region_name=boto_config.get('region_name'))
    return boto_config['session']


def get_boto_resource(config, service_name):
    boto_config = config['boto']
    session = _get_boto_session(boto_config)
    resource = session.resource(
        service_name,
        aws_access_key_id=boto_config['aws_access_key_id'],
        aws_secret_access_key=boto_config['aws_secret_access_key']
    )
    return resource


def get_boto_client(config, service_name):
    boto_config = config['boto']
    session = _get_boto_session(boto_config)
    client = session.client(
        service_name,
        aws_access_key_id=boto_config['aws_access_key_id'],
        aws_secret_access_key=boto_config['aws_secret_access_key']
    )

    return client


def tropo_to_string(snippet, indent=4, sort_keys=True, separators=(',', ': ')):
    """
    Returns the json representation of any troposphere object/template
    """
    return json.dumps(snippet, cls=t.awsencode, indent=indent, sort_keys=sort_keys, separators=separators)


def get_template_from_s3(config, template_resource_path):
    """
    Given an s3 resource path, download the template and return the json dictionary
    """
    # Download the template from s3 to a temp directory
    file_path = os.path.join(tempfile.mkdtemp(), 'downloaded_template.json')
    s3_bucket = config.get('template').get('s3_bucket')
    get_boto_client(config, "s3").download_file(s3_bucket, template_resource_path, file_path)

    # Parse the template as json and return the dictionary
    return res.load_json_file(file_path)


def get_stack_params_from_parent_template(parent_template_contents, stack_name):
    """
    This function gets all the deployment parameters used for a given stack from another deployment and returns them
    @param parent_template_contents - The loaded json contents of the parent template -- i.e., from utility.get_template_from_s3()
    @param stack_name - The name of the stack to search for in the parent template
    Returns a dictionary of stack parameters to be used with a new deployment
    """
    # Retrieve the child stack from the template
    stack_reference = parent_template_contents.get('Resources').get(stack_name)

    # If the stack is not found in the parent template, return None
    if not stack_reference:
        return None

    # Otherwise return the parameters dictionary that the stack was deployed with
    return stack_reference.get('Properties').get('Parameters')


def get_stack_depends_on_from_parent_template(parent_template_contents, stack_name):
    """
    This function gets the DependsOn attribute used for a given stack from another deployment and returns it
    @param parent_template_contents - The loaded json contents of the parent template -- i.e., from utility.get_template_from_s3()
    @param stack_name - The name of the stack to search for in the parent template
    Returns the DependsOn list to be used with a new deployment
    """
    # Retrieve the child stack from the template
    stack_reference = parent_template_contents.get('Resources').get(stack_name)

    # If the stack is not found in the parent template, return None
    if not stack_reference:
        return None

    # Otherwise return the DependsOn list that the stack was deployed with
    return stack_reference.get('DependsOn')


def get_template_s3_resource_path(prefix, template_name, include_timestamp=True):
    """
    Constructs s3 resource path for provided template name
    :param prefix: S3 base path (marts after url port and hostname)
    :param template_name: File name minus '.template' suffix and any timestamp portion
    :param include_timestamp: Indicates whether to include the current time in the file name
    :return string: Url of S3 file
    """
    if include_timestamp:
        key_serial = str(int(time.time()))
        template_name += "." + key_serial

    return "%s/%s.template" % (prefix, template_name)


def get_template_s3_url(bucket_name, resource_path):
    """
    Constructs S3 URL from bucket name and resource path.
    :param bucket_name: S3 bucket name
    :param prefix string: S3 path prefix
    :return string: S3 Url of cloudformation templates
    """
    return 'https://%s.s3.amazonaws.com/%s' % (bucket_name, resource_path)



from . import utility
import json
import time
import re

TERMINAL_STATES = [
    'CREATE_COMPLETE',
    'UPDATE_COMPLETE',
    'UPDATE_ROLLBACK_COMPLETE',
    'CREATE_FAILED',
    'UPDATE_FAILED',
    'UPDATE_ROLLBACK_FAILED',
]


class StackMonitor(object):

    def __init__(self, env_name):
        self.stack_event_handlers = []
        self.env_name = env_name

    def setup_stack_monitor(self, config):
        # Topic and queue names are randomly generated so there's no chance of picking up messages from a previous runs
        name = self.env_name + '_' + time.strftime("%Y%m%d-%H%M%S") + '_' + utility.random_string(5)

        # Creating a topic is idempotent, so if it already exists then we will just get the topic returned.
        sns = utility.get_boto_resource(config, 'sns')
        topic_arn = sns.create_topic(Name=name).arn

        # Creating a queue is idempotent, so if it already exists then we will just get the queue returned.
        sqs = utility.get_boto_resource(config, 'sqs')
        queue = sqs.create_queue(QueueName=name)

        queue_arn = queue.attributes['QueueArn']

        # Ensure that we are subscribed to the SNS topic
        subscribed = False
        topic = sns.Topic(topic_arn)
        for subscription in topic.subscriptions.all():
            if subscription.attributes['Endpoint'] == queue_arn:
                subscribed = True
                break

        if not subscribed:
            topic.subscribe(Protocol='sqs', Endpoint=queue_arn)

        # Set up a policy to allow SNS access to the queue
        if 'Policy' in queue.attributes:
            policy = json.loads(queue.attributes['Policy'])
        else:
            policy = {'Version': '2008-10-17'}

        if 'Statement' not in policy:
            statement = {
                "Sid": "sqs-access",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "SQS:SendMessage",
                "Resource": "<SQS QUEUE ARN>",
                "Condition": {"StringLike": {"aws:SourceArn": "<SNS TOPIC ARN>"}}
            }
            statement['Resource'] = queue_arn
            statement['Condition']['StringLike']['aws:SourceArn'] = topic_arn
            policy['Statement'] = [statement]

            queue.set_attributes(Attributes={
                'Policy': json.dumps(policy)
            })

        return topic, queue

    def has_handlers(self):
        return len(self.stack_event_handlers) > 0

    def add_handler(self, handler):
        self.stack_event_handlers.append(handler)

    def cleanup_stack_monitor(self, topic, queue):
        if topic:
            topic.delete()
        if queue:
            queue.delete()

    def start_stack_monitor(self, queue, stack_name, debug=False):

        # Process messages by printing out body and optional author name
        poll_timeout = 3600  # an hour
        poll_interval = 5
        start_time = time.time()
        time.clock()
        elapsed = 0
        is_stack_running = True

        while elapsed < poll_timeout and is_stack_running and len(self.stack_event_handlers) > 0:

            elapsed = time.time() - start_time

            msgs = queue.receive_messages(WaitTimeSeconds=poll_interval, MaxNumberOfMessages=10)
            # print 'grabbed batch of %s' % len(msgs)

            for raw_msg in msgs:
                parsed_msg = json.loads(raw_msg.body)
                msg_body = parsed_msg['Message']

                # parse k='val' into a dict
                parsed_msg = {k: v.strip("'") for k, v in re.findall(r"(\S+)=('.*?'|\S+)", msg_body, flags=re.DOTALL)}

                # remember the most interesting outputs
                data = {
                    "status": parsed_msg.get('ResourceStatus'),
                    "type": parsed_msg.get('ResourceType'),
                    "name": parsed_msg.get('LogicalResourceId'),
                    "id": parsed_msg.get('PhysicalResourceId'),
                    "reason": parsed_msg.get('ResourceStatusReason'),
                    "props": parsed_msg.get('ResourceProperties')
                }

                # attempt to parse the properties
                if data.get('props'):
                    try:
                        data['props'] = json.loads(data['props'])
                    except (ValueError, TypeError):
                        print("\nFailed to parse properties for event:\n{}\n".format(data['props']))
                        print("\nRaw Message Body:\n{}\n".format(msg_body))
                        print("\nParsed Message:\n{}\n".format(parsed_msg))
                        pass

                if debug:
                    print("New Stack Event ------------\n %s %s %s \n %s \n" % (data['status'], data['type'], data['name'], data['reason']))
                else:
                    pass

                # clear the message
                raw_msg.delete()

                # process handlers
                handlers_to_remove = []
                for handler in self.stack_event_handlers:
                    if handler.stack_event_hook_wrapper(data):
                        handlers_to_remove.append(handler)

                # once a handlers job is done no need to keep checking for more events
                for handler in handlers_to_remove:
                    self.stack_event_handlers.remove(handler)

                # Finally test for the termination condition
                if data['type'] == "AWS::CloudFormation::Stack" \
                        and data['name'] == stack_name \
                        and data['status'] in TERMINAL_STATES:
                    is_stack_running = False
                    # print 'termination condition found!'
